# -*- coding: utf-8 -*-

import collections
from typing import List, Tuple, Union

from requests import Response

from TM1py.Exceptions.Exceptions import TM1pyRestException
from TM1py.Objects import View
from TM1py.Objects.MDXView import MDXView
from TM1py.Objects.NativeView import NativeView
from TM1py.Services.ObjectService import ObjectService
from TM1py.Services.RestService import RestService
from TM1py.Utils import format_url


class ViewService(ObjectService):
    """ Service to handle Object Updates for cube views (NativeViews and MDXViews)
    
    """

    def __init__(self, rest: RestService):
        super().__init__(rest)

    def create(self, view: Union[MDXView, NativeView], private: bool = False, **kwargs) -> Response:
        """ create a new view on TM1 Server

        :param view: instance of subclass of TM1py.View (TM1py.NativeView or TM1py.MDXView)
        :param private: boolean

        :return: Response
        """
        view_type = "PrivateViews" if private else "Views"
        url = format_url("/api/v1/Cubes('{}')/{}", view.cube, view_type)
        return self._rest.POST(url, view.body, **kwargs)

    def exists(self, cube_name: str, view_name: str, private: bool = None, **kwargs):
        """ Checks if view exists as private, public or both

        :param cube_name:  string, name of the cube
        :param view_name: string, name of the view
        :param private: boolean, if None: check for private and public

        :return boolean tuple
        """
        url_template = "/api/v1/Cubes('{}')/{}('{}')"
        if private is not None:
            url = format_url(url_template, cube_name, "PrivateViews" if private else "Views", view_name)
            return self._exists(url, **kwargs)

        view_types = collections.OrderedDict()
        view_types['PrivateViews'] = False
        view_types['Views'] = False
        for view_type in view_types:
            try:
                url = format_url(url_template, cube_name, view_type, view_name)
                self._rest.GET(url, **kwargs)
                view_types[view_type] = True
            except TM1pyRestException as e:
                if e.status_code != 404:
                    raise e
        return tuple(view_types.values())

    def get(self, cube_name: str, view_name: str, private: bool = False, **kwargs) -> View:
        view_type = "PrivateViews" if private else "Views"
        url = format_url("/api/v1/Cubes('{}')/{}('{}')?$expand=*", cube_name, view_type, view_name)
        response = self._rest.GET(url, **kwargs)
        view_as_dict = response.json()
        if "MDX" in view_as_dict:
            return MDXView(cube_name=cube_name, view_name=view_name, MDX=view_as_dict["MDX"])
        else:
            return self.get_native_view(cube_name=cube_name, view_name=view_name, private=private)

    def get_native_view(self, cube_name: str, view_name: str, private=False, **kwargs) -> NativeView:
        """ Get a NativeView from TM1 Server

        :param cube_name:  string, name of the cube
        :param view_name:  string, name of the native view
        :param private:    boolean

        :return: instance of TM1py.NativeView
        """
        view_type = "PrivateViews" if private else "Views"
        url = format_url(
            "/api/v1/Cubes('{}')/{}('{}')?$expand="
            "tm1.NativeView/Rows/Subset($expand=Hierarchy($select=Name;"
            "$expand=Dimension($select=Name)),Elements($select=Name);"
            "$select=Expression,UniqueName,Name, Alias),  "
            "tm1.NativeView/Columns/Subset($expand=Hierarchy($select=Name;"
            "$expand=Dimension($select=Name)),Elements($select=Name);"
            "$select=Expression,UniqueName,Name,Alias), "
            "tm1.NativeView/Titles/Subset($expand=Hierarchy($select=Name;"
            "$expand=Dimension($select=Name)),Elements($select=Name);"
            "$select=Expression,UniqueName,Name,Alias), "
            "tm1.NativeView/Titles/Selected($select=Name)",
            cube_name, view_type, view_name)
        response = self._rest.GET(url, **kwargs)
        native_view = NativeView.from_json(response.text, cube_name)
        return native_view

    def get_mdx_view(self, cube_name: str, view_name: str, private: bool = False, **kwargs) -> MDXView:
        """ Get an MDXView from TM1 Server

        :param cube_name: String, name of the cube
        :param view_name: String, name of the MDX view
        :param private: boolean

        :return: instance of TM1py.MDXView
        """
        view_type = 'PrivateViews' if private else 'Views'
        url = format_url("/api/v1/Cubes('{}')/{}('{}')?$expand=*", cube_name, view_type, view_name)
        response = self._rest.GET(url, **kwargs)
        mdx_view = MDXView.from_json(view_as_json=response.text)
        return mdx_view

    def get_all(self, cube_name: str, include_elements: bool = True, **kwargs) -> Tuple[List[View], List[View]]:
        """ Get all public and private views from cube.
        :param cube_name: String, name of the cube.
        :param include_elements: false to return view details without elements, faster
        :return: 2 Lists of TM1py.View instances: private views, public views
        """

        element_filter = ";$top=0" if not include_elements else ""

        private_views, public_views = [], []
        for view_type in ('PrivateViews', 'Views'):
            url = format_url(
                "/api/v1/Cubes('{}')/{}?$expand="
                "tm1.NativeView/Rows/Subset($expand=Hierarchy($select=Name;"
                "$expand=Dimension($select=Name)),Elements($select=Name{});"
                "$select=Expression,UniqueName,Name, Alias),  "
                "tm1.NativeView/Columns/Subset($expand=Hierarchy($select=Name;"
                "$expand=Dimension($select=Name)),Elements($select=Name{});"
                "$select=Expression,UniqueName,Name,Alias), "
                "tm1.NativeView/Titles/Subset($expand=Hierarchy($select=Name;"
                "$expand=Dimension($select=Name)),Elements($select=Name{});"
                "$select=Expression,UniqueName,Name,Alias), "
                "tm1.NativeView/Titles/Selected($select=Name)",
                cube_name, view_type, element_filter, element_filter, element_filter)
            response = self._rest.GET(url, **kwargs)
            response_as_list = response.json()['value']
            for view_as_dict in response_as_list:
                if view_as_dict['@odata.type'] == '#ibm.tm1.api.v1.MDXView':
                    view = MDXView.from_dict(view_as_dict, cube_name)
                else:
                    view = NativeView.from_dict(view_as_dict, cube_name)
                if view_type == "PrivateViews":
                    private_views.append(view)
                else:
                    public_views.append(view)
        return private_views, public_views

    def get_all_names(self, cube_name: str, **kwargs) -> Tuple[List[str], List[str]]:
        """
        
        :param cube_name: 
        :return: 
        """
        private_views, public_views = [], []
        for view_type in ('PrivateViews', 'Views'):
            url = format_url("/api/v1/Cubes('{}')/{}?$select=Name", cube_name, view_type)
            response = self._rest.GET(url, **kwargs)
            response_as_list = response.json()['value']

            for view in response_as_list:
                if view_type == "PrivateViews":
                    private_views.append(view['Name'])
                else:
                    public_views.append(view['Name'])

        return private_views, public_views

    def update(self, view: Union[MDXView, NativeView], private: bool = False, **kwargs) -> Response:
        """ Update an existing view

        :param view: instance of TM1py.NativeView or TM1py.MDXView
        :param private: boolean
        :return: response
        """
        view_type = 'PrivateViews' if private else 'Views'
        url = format_url("/api/v1/Cubes('{}')/{}('{}')", view.cube, view_type, view.name)
        response = self._rest.PATCH(url, view.body, **kwargs)
        return response

    def update_or_create(self, view: Union[MDXView, NativeView], private: bool = False, **kwargs) -> Response:
        """ update if exists, else create

        :param view:
        :param private:
        :param kwargs:
        :return:
        """
        if self.exists(view.cube, view.name, private=private, **kwargs):
            return self.update(view, private=private, **kwargs)

        return self.create(view, private=private, **kwargs)

    def delete(self, cube_name: str, view_name: str, private: bool = False, **kwargs) -> Response:
        """ Delete an existing view (MDXView or NativeView) on the TM1 Server

        :param cube_name: String, name of the cube
        :param view_name: String, name of the view
        :param private: Boolean

        :return: String, the response
        """
        view_type = 'PrivateViews' if private else 'Views'
        url = format_url("/api/v1/Cubes('{}')/{}('{}')", cube_name, view_type, view_name)
        response = self._rest.DELETE(url, **kwargs)
        return response

    def search_subset_in_native_views(self, dimension_name: str = None, subset_name: str = None, cube_name: str = None,
                                      include_elements: bool = False, **kwargs) -> Tuple[List[View], List[View]]:
        """ Get all public and private native views that utilize specified dimension subset

        :param dimension_name: string, valid dimension name with subset to query
        :param subset_name: string, valid subset name to search for in views
        :param cube_name: str, optionally specify cube to search, otherwise will search all cubes
        :param include_elements: false to return view details without elements, faster
        :return: 2 Lists of TM1py.View instances: private views, public views
        """

        dimension_name = dimension_name.lower().replace(' ', '')
        subset_name = subset_name.lower().replace(' ', '')

        element_filter = ";$top=0" if not include_elements else ""
        if cube_name:
            base_url = format_url(
                "/api/v1/Cubes?$select=Name&$filter=replace(tolower(Name),' ', '') eq '{}'",
                cube_name.lower().replace(' ', ''))
        else:
            base_url = "/api/v1/Cubes?$select=Name"

        private_views, public_views = [], []
        for view_type in ('PrivateViews', 'Views'):
            url = base_url + format_url(
                "&$expand={}($filter=isof(tm1.NativeView) and"
                "("
                "(tm1.NativeView/Rows/any (r: replace(tolower(r/Subset/Name), ' ', '') eq '{}' "
                "and replace(tolower(r/Subset/Hierarchy/Dimension/Name), ' ', '') eq '{}'))"
                "or"
                "(tm1.NativeView/Columns/any (c: replace(tolower(c/Subset/Name), ' ', '') eq '{}' "
                "and replace(tolower(c/Subset/Hierarchy/Dimension/Name), ' ', '') eq '{}')) "
                "or"
                "(tm1.NativeView/Titles/any (t: replace(tolower(t/Subset/Name), ' ', '') eq '{}' "
                "and replace(tolower(t/Subset/Hierarchy/Dimension/Name), ' ', '') eq '{}'))"
                ");"
                "$expand=tm1.NativeView/Rows/Subset($expand=Hierarchy($select=Name;"
                "$expand=Dimension($select=Name)),Elements($select=Name{});"
                "$select=Expression,UniqueName,Name, Alias),  "
                "tm1.NativeView/Columns/Subset($expand=Hierarchy($select=Name;"
                "$expand=Dimension($select=Name)),Elements($select=Name{});"
                "$select=Expression,UniqueName,Name,Alias), "
                "tm1.NativeView/Titles/Subset($expand=Hierarchy($select=Name;"
                "$expand=Dimension($select=Name)),Elements($select=Name{});"
                "$select=Expression,UniqueName,Name,Alias), "
                "tm1.NativeView/Titles/Selected($select=Name))",
                view_type, subset_name, dimension_name, subset_name, dimension_name,
                subset_name, dimension_name, element_filter, element_filter, element_filter
            )

            response = self._rest.GET(url, **kwargs)
            response_as_list = response.json()['value']
            for cube in response_as_list:
                for view_as_dict in cube[view_type]:
                    view = NativeView.from_dict(view_as_dict, cube['Name'])
                    if view_type == "PrivateViews":
                        private_views.append(view)
                    else:
                        public_views.append(view)

        return private_views, public_views



# -*- coding: utf-8 -*-
from abc import abstractmethod

from TM1py.Objects.TM1Object import TM1Object


class View(TM1Object):
    """ Abstraction of TM1 View
        serves as a parentclass for TM1py.Objects.MDXView and TM1py.Objects.NativeView

    """

    def __init__(self, cube: str, name: str):
        self._cube = cube
        self._name = name

    @abstractmethod
    def body(self) -> str:
        pass

    @property
    def cube(self) -> str:
        return self._cube

    @property
    def name(self) -> str:
        return self._name

    @cube.setter
    def cube(self, value: str):
        self._cube = value

    @name.setter
    def name(self, value: str):
        self._name = value

    @property
    def mdx(self):
        raise NotImplementedError



# -*- coding: utf-8 -*-


class GitCommit:
    """ Abstraction of Git Commit
    """

    def __init__(self, commit_id: str, summary: str, author: str):
        """ Initialize GitCommit object
        :param commit_id: id of the commit
        :param summary: commit message
        :param author: the author of the commit
        """
        self._commit_id = commit_id
        self._summary = summary
        self._author = author

    @property
    def commit_id(self) -> str:
        return self._commit_id

    @property
    def summary(self) -> str:
        return self._summary

    @property
    def author(self) -> str:
        return self._author



# -*- coding: utf-8 -*-
from typing import Dict

from TM1py.Objects.GitCommit import GitCommit
from TM1py.Objects.GitRemote import GitRemote


class Git:
    """ Abstraction of Git object
    """

    def __init__(self, url: str, deployment: str, force: bool, deployed_commit: GitCommit, remote: GitRemote,
                 config: dict = None):
        """ Initialize GIT object
        :param url: file or http(s) path to GIT repository
        :param deployment: name of selected deployment group
        :param force: whether or not Git context was forced
        :param deployed_commit: GitCommit object of the currently deployed commit
        :param remote: GitRemote object of the current remote
        :param config: Dictionary containing git configuration parameters

        """
        self._url = url
        self._deployment = deployment
        self._force = force
        self._deployed_commit = deployed_commit
        self._remote = remote
        self._config = config

    @property
    def url(self) -> str:
        return self._url

    @property
    def force(self) -> bool:
        return self._force

    @property
    def config(self) -> dict:
        return self._config

    @property
    def deployment(self) -> str:
        return self._deployment

    @property
    def deployed_commit(self) -> GitCommit:
        return self._deployed_commit

    @property
    def remote(self) -> GitRemote:
        return self._remote

    @classmethod
    def from_dict(cls, json_response: Dict) -> 'Git':
        deployed_commit = GitCommit(
            commit_id=json_response["DeployedCommit"].get("ID"),
            summary=json_response["DeployedCommit"].get("Summary"),
            author=json_response["DeployedCommit"].get("Author")
        )

        remote = GitRemote(
            connected=json_response["Remote"].get("Connected"),
            branches=json_response["Remote"].get("Branches"),
            tags=json_response["Remote"].get("Tags"),
        )

        git = Git(
            url=json_response["URL"],
            deployment=json_response["Deployment"],
            force=json_response["Deployment"],
            deployed_commit=deployed_commit,
            remote=remote)

        return git



"""
A python module for TM1.

https://github.com/cubewise-code/TM1py

TM1py wraps the TM1 REST API into concise Python classes and Services that simplify TM1 interactions from python.

Usage:
>>> with RestService(address='', port=8001, user='admin', password='apple', ssl=False) as tm1_rest:
>>>     subset_service = SubsetService(tm1_rest)
>>>     subset = Subset(dimension_name='Month', subset_name='Q1', elements=['Jan', 'Feb', 'Mar'])
>>>     subset_service.create(subset, private=True)

"""

# __init__ can hoist attributes from submodules into higher namespaces for convenience

from TM1py.Objects.Annotation import Annotation
from TM1py.Objects.Application import Application
from TM1py.Objects.Axis import ViewAxisSelection, ViewTitleSelection
from TM1py.Objects.Chore import Chore
from TM1py.Objects.ChoreFrequency import ChoreFrequency
from TM1py.Objects.ChoreStartTime import ChoreStartTime
from TM1py.Objects.ChoreTask import ChoreTask
from TM1py.Objects.Cube import Cube
from TM1py.Objects.Dimension import Dimension
from TM1py.Objects.Element import Element
from TM1py.Objects.ElementAttribute import ElementAttribute
from TM1py.Objects.Git import Git
from TM1py.Objects.GitCommit import GitCommit
from TM1py.Objects.GitPlan import GitPlan
from TM1py.Objects.GitRemote import GitRemote
from TM1py.Objects.Hierarchy import Hierarchy
from TM1py.Objects.MDXView import MDXView
from TM1py.Objects.NativeView import NativeView
from TM1py.Objects.Process import Process
from TM1py.Objects.Rules import Rules
from TM1py.Objects.Sandbox import Sandbox
from TM1py.Objects.Server import Server
from TM1py.Objects.Subset import Subset, AnonymousSubset
from TM1py.Objects.User import User
from TM1py.Objects.View import View
from TM1py.Services.AnnotationService import AnnotationService
from TM1py.Services.ApplicationService import ApplicationService
from TM1py.Services.CellService import CellService
from TM1py.Services.ChoreService import ChoreService
from TM1py.Services.CubeService import CubeService
from TM1py.Services.DimensionService import DimensionService
from TM1py.Services.ElementService import ElementService
from TM1py.Services.GitService import GitService
from TM1py.Services.HierarchyService import HierarchyService
from TM1py.Services.MonitoringService import MonitoringService
from TM1py.Services.ObjectService import ObjectService
from TM1py.Services.PowerBiService import PowerBiService
from TM1py.Services.ProcessService import ProcessService
from TM1py.Services.RestService import RestService
from TM1py.Services.SandboxService import SandboxService
from TM1py.Services.SecurityService import SecurityService
from TM1py.Services.ServerService import ServerService
from TM1py.Services.SubsetService import SubsetService
from TM1py.Services.TM1Service import TM1Service
from TM1py.Services.ViewService import ViewService
from TM1py.Utils import Utils



from collections.abc import Iterable

from TM1py.Services import CellService
from TM1py.Services import ElementService
from TM1py.Utils import require_pandas

try:
    import pandas as pd

    _has_pandas = True
except ImportError:
    _has_pandas = False


class PowerBiService:
    def __init__(self, tm1_rest):
        """

        :param tm1_rest: instance of RestService
        """
        self._tm1_rest = tm1_rest
        self.cells = CellService(tm1_rest)
        self.elements = ElementService(tm1_rest)

    @require_pandas
    def execute_mdx(self, mdx, **kwargs) -> 'pd.DataFrame':
        return self.cells.execute_mdx_dataframe_shaped(mdx, **kwargs)

    @require_pandas
    def execute_view(self, cube_name, view_name, private, **kwargs) -> 'pd.DataFrame':
        return self.cells.execute_view_dataframe_shaped(cube_name, view_name, private, **kwargs)

    @require_pandas
    def get_member_properties(self, dimension_name: str, hierarchy_name: str, member_selection: Iterable = None,
                              skip_consolidations: bool = True, attributes: Iterable = None,
                              skip_parents: bool = False, level_names=None,
                              parent_attribute: str = None) -> 'pd.DataFrame':
        """

        :param dimension_name: Name of the dimension
        :param hierarchy_name: Name of the hierarchy in the dimension
        :param member_selection: Selection of members. Iterable or valid MDX string
        :param skip_consolidations: Boolean flag to skip consolidations
        :param attributes: Selection of attributes. Iterable. If None retrieve all.
        :param level_names: List of labels for parent columns. If None use level names from TM1.
        :param skip_parents: Boolean Flag to skip parent columns.
        :param parent_attribute: Attribute to be displayed in parent columns. If None, parent name is used.
        :return: pandas DataFrame
        """
        if not member_selection:
            member_selection = f"{{ [{dimension_name}].[{hierarchy_name}].Members }}"
            if skip_consolidations:
                member_selection = f"{{ Tm1FilterByLevel({member_selection}, 0) }}"

        if not isinstance(member_selection, str):
            if isinstance(member_selection, Iterable):
                member_selection = "{" + ",".join(f"[{dimension_name}].[{member}]" for member in member_selection) + "}"
            else:
                raise ValueError("Argument 'element_selection' must be None or str")

        if not self.elements.attribute_cube_exists(dimension_name):
            raise RuntimeError(self.elements.ELEMENT_ATTRIBUTES_PREFIX + dimension_name + " cube must exist")

        members = [tupl[0] for tupl in self.elements.execute_set_mdx(
            mdx=member_selection,
            element_properties=None,
            member_properties=("Name", "UniqueName"),
            parent_properties=None)]

        element_types = self.elements.get_element_types(
            dimension_name=dimension_name,
            hierarchy_name=hierarchy_name,
            skip_consolidations=skip_consolidations)

        df = pd.DataFrame(
            data=[(member["Name"], element_types[member["Name"]])
                  for member
                  in members
                  if member["Name"] in element_types],
            dtype=str,
            columns=[dimension_name, 'Type'])

        calculated_members_definition = list()
        calculated_members_selection = list()
        if not skip_parents:
            levels = self.elements.get_levels_count(dimension_name, hierarchy_name)

            # potential custom parent names
            if not level_names:
                level_names = self.elements.get_level_names(dimension_name, hierarchy_name, descending=True)

            for parent in range(1, levels, 1):
                name_or_attribute = f"Properties('{parent_attribute}')" if parent_attribute else "Name"
                member = f"""
                    MEMBER [{self.elements.ELEMENT_ATTRIBUTES_PREFIX + dimension_name}].[{level_names[parent]}] 
                    AS [{dimension_name}].CurrentMember.{'Parent.' * parent}{name_or_attribute}
                    """
                calculated_members_definition.append(member)

                calculated_members_selection.append(
                    f"[{self.elements.ELEMENT_ATTRIBUTES_PREFIX + dimension_name}].[{level_names[parent]}]")

        if attributes is None:
            column_selection = "{Tm1SubsetAll([" + self.elements.ELEMENT_ATTRIBUTES_PREFIX + dimension_name + "])}"
        else:
            column_selection = "{" + ",".join(
                "[" + self.elements.ELEMENT_ATTRIBUTES_PREFIX + dimension_name + "].[" + attribute + "]"
                for attribute
                in attributes) + "}"

        if calculated_members_selection:
            column_selection = column_selection + " + {" + ",".join(calculated_members_selection) + "}"
        member_selection = ",".join(
            member["UniqueName"]
            for member
            in members)

        mdx_with_block = ""
        if calculated_members_definition:
            mdx_with_block = "WITH " + " ".join(calculated_members_definition)

        mdx = f"""
        {mdx_with_block}
        SELECT
        {{ {member_selection} }} ON ROWS,
        {{ {column_selection} }} ON COLUMNS
        FROM [{self.elements.ELEMENT_ATTRIBUTES_PREFIX + dimension_name}]  
        """

        df_data = self.execute_mdx(mdx)

        # override hierarchy name
        df_data.rename(columns={hierarchy_name:dimension_name},inplace=True)
        
        # shift levels to right hand side
        if not skip_parents:
            # skip max level (= leaves)
            level_names = level_names[1:]
            # iterative approach
            for _ in level_names:
                rows_to_shift = df_data[df_data[level_names[-1]] == ''].index
                if rows_to_shift.empty:
                    break
                df_data.iloc[rows_to_shift, -len(level_names):] = df_data.iloc[rows_to_shift, -len(level_names):].shift(
                    1, axis=1)

            df_data.iloc[:, -len(level_names):] = df_data.iloc[:, -len(level_names):].fillna('')

        return pd.merge(df, df_data, on=dimension_name).drop_duplicates()



# -*- coding: utf-8 -*-

import collections
import json
from enum import Enum
from typing import Union, Dict, List

from TM1py.Objects.TM1Object import TM1Object
from TM1py.Utils import case_and_space_insensitive_equals


class Element(TM1Object):
    """ Abstraction of TM1 Element

    """
    ELEMENT_ATTRIBUTES_PREFIX = "}ElementAttributes_"

    class Types(Enum):
        NUMERIC = 1
        STRING = 2
        CONSOLIDATED = 3

        def __str__(self):
            return self.name.capitalize()

        @classmethod
        def _missing_(cls, value: str):
            for member in cls:
                if member.name.lower() == value.replace(" ", "").lower():
                    return member
            # default
            raise ValueError(f"Invalid element type: '{value}'")

    def __init__(self, name, element_type: Union[Types, str], attributes: List[str] = None, unique_name: str = None,
                 index: int = None):
        self._name = name
        self._unique_name = unique_name
        self._index = index
        self._element_type = None
        self.element_type = element_type
        self._attributes = attributes

    @staticmethod
    def from_dict(element_as_dict: Dict) -> 'Element':
        return Element(name=element_as_dict['Name'],
                       unique_name=element_as_dict.get('UniqueName', None),
                       index=element_as_dict.get('Index', None),
                       element_type=element_as_dict['Type'],
                       attributes=element_as_dict.get('Attributes', None))

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    @property
    def unique_name(self) -> str:
        return self._unique_name

    @property
    def index(self) -> int:
        return self._index

    @property
    def element_attributes(self) -> List[str]:
        return self._attributes

    @property
    def element_type(self) -> Types:
        return self._element_type

    @element_type.setter
    def element_type(self, value: Union[Types, str]):
        self._element_type = Element.Types(value)

    @property
    def body(self) -> str:
        return json.dumps(self._construct_body())

    @property
    def body_as_dict(self) -> Dict:
        return self._construct_body()

    def _construct_body(self) -> Dict:
        body_as_dict = collections.OrderedDict()
        body_as_dict['Name'] = self._name
        body_as_dict['Type'] = str(self._element_type)
        return body_as_dict

    def __eq__(self, other: 'Element'):
        return all([
            isinstance(other, Element),
            case_and_space_insensitive_equals(self.name, other.name),
            self.element_type == other.element_type])

    def __hash__(self):
        return super().__hash__()



# -*- coding: utf-8 -*-

import collections
import json
from typing import List, Dict, Iterable, Optional, Tuple, Union, Set

from TM1py.Objects.Element import Element
from TM1py.Objects.ElementAttribute import ElementAttribute
from TM1py.Objects.TM1Object import TM1Object
from TM1py.Utils.Utils import CaseAndSpaceInsensitiveDict, CaseAndSpaceInsensitiveTuplesDict, lower_and_drop_spaces, \
    case_and_space_insensitive_equals


class Hierarchy(TM1Object):
    """ Abstraction of TM1 Hierarchy
        Requires reference to a Dimension

        Elements modeled as a Dictionary where key is the element name and value an instance of TM1py.Element
        {
            'US': instance of TM1py.Element,
            'CN': instance of TM1py.Element,
            'AU': instance of TM1py.Element
        }

        ElementAttributes of type TM1py.Objects.ElementAttribute

        Edges are represented as a TM1py.Utils.CaseAndSpaceInsensitiveTupleDict: 
        {
            (parent1, component1) : 10,
            (parent1, component2) : 30
        }

        Subsets is list of type TM1py.Subset

    """

    def __init__(
            self,
            name: str,
            dimension_name: str,
            elements: Optional[Iterable['Element']] = None,
            element_attributes: Optional[Iterable['ElementAttribute']] = None,
            edges: Optional['Dict'] = None,
            subsets: Optional[Iterable[str]] = None,
            structure: Optional[int] = None,
            default_member: Optional[str] = None):

        self._name = name
        self._dimension_name = None
        self.dimension_name = dimension_name
        self._elements: Dict[str, Element] = CaseAndSpaceInsensitiveDict()
        if elements:
            for elem in elements:
                self._elements[elem.name] = elem
        self._element_attributes = list(element_attributes) if element_attributes else []
        self._edges = CaseAndSpaceInsensitiveTuplesDict(edges) if edges else CaseAndSpaceInsensitiveTuplesDict()
        self._subsets = list(subsets) if subsets else []
        # balanced is true, false or None (in versions < TM1 11)
        self._balanced = False if not structure else structure == 0
        self._default_member = default_member

    @classmethod
    def from_dict(cls, hierarchy_as_dict: Dict, dimension_name: str = None) -> 'Hierarchy':
        # Build the Dictionary for the edges
        edges = CaseAndSpaceInsensitiveTuplesDict(
            {(edge['ParentName'], edge['ComponentName']): edge['Weight']
             for edge
             in hierarchy_as_dict['Edges']})

        if not dimension_name:
            dimension_name = hierarchy_as_dict['UniqueName'][1:hierarchy_as_dict['UniqueName'].find("].[")]

        return cls(
            name=hierarchy_as_dict['Name'],
            dimension_name=dimension_name,
            elements=[Element.from_dict(elem) for elem in hierarchy_as_dict['Elements']],
            element_attributes=[ElementAttribute(ea['Name'], ea['Type'])
                                for ea in hierarchy_as_dict.get('ElementAttributes', [])],
            edges=edges,
            subsets=[subset['Name'] for subset in hierarchy_as_dict.get('Subsets', [])],
            structure=hierarchy_as_dict['Structure'] if 'Structure' in hierarchy_as_dict else None,
            default_member=hierarchy_as_dict['DefaultMember']['Name']
            if hierarchy_as_dict.get('DefaultMember', None) else None)

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    @property
    def dimension_name(self) -> str:
        return self._dimension_name

    @dimension_name.setter
    def dimension_name(self, dimension_name: str):
        self._dimension_name = dimension_name

    @property
    def elements(self) -> Dict[str, Element]:
        return self._elements

    @property
    def element_attributes(self) -> List[ElementAttribute]:
        return self._element_attributes

    @property
    def edges(self) -> Dict[Tuple[str], Element]:
        return self._edges

    @property
    def subsets(self) -> List[str]:
        return self._subsets

    @property
    def balanced(self) -> bool:
        return self._balanced

    @property
    def default_member(self) -> str:
        return self._default_member

    @property
    def body(self) -> str:
        return json.dumps(self._construct_body())

    @property
    def body_as_dict(self) -> Dict:
        return self._construct_body()

    def contains_element(self, element_name: str) -> bool:
        return element_name in self._elements

    def get_element(self, element_name: str) -> Element:
        if element_name in self._elements:
            return self._elements[element_name]
        else:
            raise ValueError("Element: {} not found in Hierarchy: {}".format(element_name, self.name))

    def get_ancestors(self, element_name: str, recursive: bool = False) -> Set[Element]:
        ancestors = set()

        for (parent, component) in self._edges:
            if not case_and_space_insensitive_equals(component, element_name):
                continue

            ancestor: Element = self.elements[parent]
            ancestors.add(ancestor)

            if recursive:
                ancestors = ancestors.union(self.get_ancestors(ancestor.name, True))
        return ancestors

    def get_descendants(self, element_name: str, recursive: bool = False, leaves_only=False) -> Set[Element]:
        descendants = set()

        for (parent, component) in self._edges:
            if not case_and_space_insensitive_equals(parent, element_name):
                continue

            descendant: Element = self.elements[component]
            if not leaves_only:
                descendants.add(descendant)
            else:
                if descendant.element_type == Element.Types.NUMERIC:
                    descendants.add(descendant)

            if recursive and descendant.element_type == Element.Types.CONSOLIDATED:
                descendants = descendants.union(self.get_descendants(descendant.name, True))
        return descendants

    def get_descendant_edges(self, element_name: str, recursive: bool = False) -> Dict:
        descendant_edges = dict()

        for (parent, component), weight in self._edges.items():
            if not case_and_space_insensitive_equals(parent, element_name):
                continue

            descendant_edges[parent, component] = weight
            descendant: Element = self.elements[component]

            if recursive and descendant.element_type == Element.Types.CONSOLIDATED:
                descendant_edges.update(self.get_descendant_edges(descendant.name, True))

        return descendant_edges
        
    def add_element(self, element_name: str, element_type: Union[str, Element.Types]):
        if element_name in self._elements:
            raise ValueError("Element name must be unique")

        self._elements[element_name] = Element(name=element_name, element_type=element_type)

    def add_component(self, parent_name: str, component_name: str, weight: int):
        if parent_name not in self._elements:
            raise ValueError(f"Parent '{parent_name}' does not exist in hierarchy")
        if self._elements[parent_name].element_type != Element.Types.CONSOLIDATED:
            raise ValueError(f"Parent '{parent_name}' is not of type 'Consolidated'")

        if component_name not in self.elements:
            self.add_element(component_name, 'Numeric')
        elif self._elements[component_name].element_type == Element.Types.STRING:
            raise ValueError(f"Component '{component_name}' must not be of type 'String'")

        self.add_edge(parent_name, component_name, weight)

    def update_element(self, element_name: str, element_type: Union[str, Element.Types]):
        self._elements[element_name].element_type = element_type

    def remove_element(self, element_name: str):
        if element_name not in self._elements:
            return
        del self._elements[element_name]
        self.remove_edges_related_to_element(element_name=element_name)

    def remove_all_elements(self):
        self._elements = CaseAndSpaceInsensitiveDict()
        self.remove_all_edges()

    def add_edge(self, parent: str, component: str, weight: int):
        self._edges[(parent, component)] = weight

    def update_edge(self, parent: str, component: str, weight: int):
        self._edges[(parent, component)] = weight

    def remove_edge(self, parent: str, component: str):
        if (parent, component) in self.edges:
            del self.edges[(parent, component)]

    def remove_edges(self, edges: Iterable[Tuple[str, str]]):
        for edge in edges:
            self.remove_edge(*edge)

    def remove_all_edges(self):
        self._edges = CaseAndSpaceInsensitiveTuplesDict()

    def remove_edges_related_to_element(self, element_name: str):
        element_name_adjusted = lower_and_drop_spaces(element_name)
        edges_to_remove = set()
        for edge in self._edges.adjusted_keys():
            if element_name_adjusted in edge:
                edges_to_remove.add(edge)
        self.remove_edges(edges=edges_to_remove)

    def add_element_attribute(self, name: str, attribute_type: str):
        attribute = ElementAttribute(name, attribute_type)
        if attribute not in self.element_attributes:
            self.element_attributes.append(attribute)

    def remove_element_attribute(self, name: str):
        self._element_attributes = [
            element_attribute
            for element_attribute
            in self.element_attributes if not case_and_space_insensitive_equals(element_attribute.name, name)]

    def _construct_body(self, element_attributes: Optional[bool] = False) -> Dict:
        """
        With TM1 10.2.2 Hierarchy and Element Attributes can't be created in one batch
        -> https://www.ibm.com/developerworks/community/forums/html/threadTopic?id=d91f3e0e-d305-44db-ac02-2fdcbee00393
        Thus, no need to have the ElementAttribute included in the JSON

        :param element_attributes: Only include element_attributes in body if explicitly asked for
        :return:
        """

        body_as_dict = collections.OrderedDict()
        body_as_dict['Name'] = self._name
        body_as_dict['Elements'] = []
        body_as_dict['Edges'] = []

        for element in self._elements.values():
            body_as_dict['Elements'].append(element.body_as_dict)
        for edge in self._edges:
            edge_as_dict = collections.OrderedDict()
            edge_as_dict['ParentName'] = edge[0]
            edge_as_dict['ComponentName'] = edge[1]
            edge_as_dict['Weight'] = self._edges[edge]
            body_as_dict['Edges'].append(edge_as_dict)
        if element_attributes:
            body_as_dict['ElementAttributes'] = [element_attribute.body_as_dict
                                                 for element_attribute
                                                 in self._element_attributes]
        return body_as_dict

    def __iter__(self):
        return iter(self._elements.values())

    def __len__(self):
        return len(self._elements)

    def __contains__(self, item):
        return self.contains_element(item)

    def __getitem__(self, item):
        return self.get_element(item)



from TM1py.Services.AnnotationService import AnnotationService
from TM1py.Services.ApplicationService import ApplicationService
from TM1py.Services.CellService import CellService
from TM1py.Services.ChoreService import ChoreService
from TM1py.Services.CubeService import CubeService
from TM1py.Services.DimensionService import DimensionService
from TM1py.Services.ElementService import ElementService
from TM1py.Services.GitService import GitService
from TM1py.Services.HierarchyService import HierarchyService
from TM1py.Services.MonitoringService import MonitoringService
from TM1py.Services.PowerBiService import PowerBiService
from TM1py.Services.ProcessService import ProcessService
from TM1py.Services.RestService import RestService
from TM1py.Services.SandboxService import SandboxService
from TM1py.Services.SecurityService import SecurityService
from TM1py.Services.ServerService import ServerService
from TM1py.Services.SubsetService import SubsetService
from TM1py.Services.ViewService import ViewService
from TM1py.Services.GitService import GitService
from TM1py.Services.TM1Service import TM1Service



# -*- coding: utf-8 -*-

import collections
import json
import re
from typing import Optional, Dict

from TM1py.Objects.View import View
from TM1py.Utils import case_and_space_insensitive_equals


class MDXView(View):
    """ Abstraction on TM1 MDX view

        IMPORTANT. MDXViews can't be seen through the old TM1 clients (Archict, Perspectives). They do exist though!
    """

    def __init__(self, cube_name: str, view_name: str, MDX: str):
        View.__init__(self, cube_name, view_name)
        self._mdx = MDX

    @property
    def mdx(self):
        return self._mdx

    @mdx.setter
    def mdx(self, value: str):
        self._mdx = value

    @property
    def MDX(self) -> str:
        return self._mdx

    @MDX.setter
    def MDX(self, value: str):
        self._mdx = value

    @property
    def body(self) -> str:
        return self.construct_body()

    def substitute_title(self, dimension: str, hierarchy: str, element: str):
        """ dimension and hierarchy name are space sensitive!

        :param dimension:
        :param hierarchy:
        :param element:
        :return:
        """
        pattern = re.compile(r"\[" + dimension + r"\].\[" + hierarchy + r"\].\[(.*?)\]", re.IGNORECASE)
        findings = re.findall(pattern, self._mdx)

        if findings:
            self._mdx = re.sub(
                pattern=pattern,
                repl=f"[{dimension}].[{hierarchy}].[{element}]",
                string=self._mdx)
            return

        if hierarchy is None or case_and_space_insensitive_equals(dimension, hierarchy):
            pattern = re.compile(r"\[" + dimension + r"\].\[(.*?)\]", re.IGNORECASE)
            findings = re.findall(pattern, self._mdx)
            if findings:
                self._mdx = re.sub(
                    pattern=pattern,
                    repl=f"[{dimension}].[{element}]",
                    string=self._mdx)
                return

        raise ValueError(f"No selection in title with dimension: '{dimension}' and hierarchy: '{hierarchy}'")

    @classmethod
    def from_json(cls, view_as_json: str, cube_name: Optional[str] = None) -> 'MDXView':
        view_as_dict = json.loads(view_as_json)
        return cls.from_dict(view_as_dict, cube_name)

    @classmethod
    def from_dict(cls, view_as_dict: Dict, cube_name: str = None) -> 'MDXView':
        return cls(cube_name=view_as_dict['Cube']['Name'] if not cube_name else cube_name,
                   view_name=view_as_dict['Name'],
                   MDX=view_as_dict['MDX'])

    def construct_body(self) -> str:
        mdx_view_as_dict = collections.OrderedDict()
        mdx_view_as_dict['@odata.type'] = 'ibm.tm1.api.v1.MDXView'
        mdx_view_as_dict['Name'] = self._name
        mdx_view_as_dict['MDX'] = self._mdx
        return json.dumps(mdx_view_as_dict, ensure_ascii=False)



# -*- coding: utf-8 -*-

# TM1py Exceptions are defined here
from typing import Mapping, List


class TM1pyTimeout(Exception):
    def __init__(self, method: str, url: str, timeout: float):
        self.method = method
        self.url = url
        self.timeout = timeout

    def __str__(self):
        return f"Timeout after {self.timeout} seconds for '{self.method}' request with url :'{self.url}'"


class TM1pyVersionException(Exception):
    def __init__(self, function: str, required_version):
        self.function = function
        self.required_version = required_version

    def __str__(self):
        return f"Function '{self.function}' requires TM1 server version >= '{self.required_version}'"


class TM1pyNotAdminException(Exception):
    def __init__(self, function: str):
        self.function = function

    def __str__(self):
        return f"Function '{self.function}' requires admin permissions"


class TM1pyException(Exception):
    """ The default exception for TM1py

    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class TM1pyRestException(TM1pyException):
    """ Exception for failing REST operations

    """

    def __init__(self, response: str, status_code: int, reason: str, headers: Mapping):
        super(TM1pyRestException, self).__init__(response)
        self._status_code = status_code
        self._reason = reason
        self._headers = headers

    @property
    def status_code(self):
        return self._status_code

    @property
    def reason(self):
        return self._reason

    @property
    def response(self):
        return self.message

    @property
    def headers(self):
        return self._headers

    def __str__(self):
        return "Text: '{}' - Status Code: {} - Reason: '{}' - Headers: {}".format(
            self.message,
            self._status_code,
            self._reason,
            self._headers)


class TM1pyWriteFailureException(TM1pyException):

    def __init__(self, statuses: List[str], error_log_files: List[str]):
        self.statuses = statuses
        self.error_log_files = error_log_files

        message = f"All {len(self.statuses)} write operations failed. Details: {self.error_log_files}"
        super(TM1pyWriteFailureException, self).__init__(message)


class TM1pyWritePartialFailureException(TM1pyException):

    def __init__(self, statuses: List[str], error_log_files: List[str], attempts: int):
        self.statuses = statuses
        self.error_log_files = error_log_files
        self.attempts = attempts

        message = f"{len(self.statuses)} out of {self.attempts} write operations failed partially. " \
                  f"Details: {self.error_log_files}"
        super(TM1pyWritePartialFailureException, self).__init__(message)



from typing import List, Iterable
from requests import Response
import json

from TM1py.Exceptions.Exceptions import TM1pyRestException
from TM1py.Services.ObjectService import ObjectService
from TM1py.Services.RestService import RestService
from TM1py.Utils import format_url
from TM1py.Objects.Sandbox import Sandbox


class SandboxService(ObjectService):
    """ Service to handle sandboxes in TM1
    
    """

    def __init__(self, rest: RestService):
        super().__init__(rest)

    def get(self, sandbox_name: str, **kwargs) -> Sandbox:
        """ get a sandbox from TM1 Server

        :param sandbox_name: str
        :return: instance of TM1py.Sandbox
        """
        url = format_url("/api/v1/Sandboxes('{}')", sandbox_name)
        response = self._rest.GET(url=url, **kwargs)
        sandbox = Sandbox.from_json(response.text)
        return sandbox

    def get_all(self, **kwargs) -> List[Sandbox]:
        """ get all sandboxes from TM1 Server

        :return: List of TM1py.Sandbox instances
        """
        url = "/api/v1/Sandboxes?$select=Name,IncludeInSandboxDimension"
        response = self._rest.GET(url, **kwargs)
        sandboxes = [
            Sandbox.from_dict(sandbox_as_dict=sandbox)
            for sandbox in response.json()["value"]
        ]
        return sandboxes

    def get_all_names(self, **kwargs) -> List[str]:
        """ get all sandbox names

        :param kwargs:
        :return:
        """
        url = "/api/v1/Sandboxes?$select=Name"
        response = self._rest.GET(url, **kwargs)
        return [entry["Name"] for entry in response.json()["value"]]

    def create(self, sandbox: Sandbox, **kwargs) -> Response:
        """ create a new sandbox in TM1 Server

        :param sandbox: Sandbox
        :return: response
        """
        url = "/api/v1/Sandboxes"
        return self._rest.POST(url=url, data=sandbox.body, **kwargs)

    def update(self, sandbox: Sandbox, **kwargs) -> Response:
        """ update a sandbox in TM1

        :param sandbox:
        :return: response
        """
        url = format_url("/api/v1/Sandboxes('{}')", sandbox.name)
        return self._rest.PATCH(url=url, data=sandbox.body, **kwargs)

    def delete(self, sandbox_name: str, **kwargs) -> Response:
        """ delete a sandbox in TM1

        :param sandbox_name:
        :return: response
        """
        url = format_url("/api/v1/Sandboxes('{}')", sandbox_name)
        return self._rest.DELETE(url, **kwargs)

    def publish(self, sandbox_name: str, **kwargs) -> Response:
        """ publish existing sandbox to base

        :param sandbox_name: str
        :return: response
        """
        url = format_url("/api/v1/Sandboxes('{}')/tm1.Publish", sandbox_name)
        return self._rest.POST(url=url, **kwargs)

    def reset(self, sandbox_name: str, **kwargs) -> Response:
        """ reset all changes in specified sandbox

        :param sandbox_name: str
        :return: response
        """
        url = format_url("/api/v1/Sandboxes('{}')/tm1.DiscardChanges", sandbox_name)
        return self._rest.POST(url=url, **kwargs)

    def merge(
        self,
        source_sandbox_name: str,
        target_sandbox_name: str,
        clean_after: bool = False,
        **kwargs
    ) -> Response:
        """ merge one sandbox into another

        :param source_sandbox_name: str
        :param target_sandbox_name: str
        :param clean_after: bool: Reset source sandbox after merging
        :return: response
        """
        url = format_url("/api/v1/Sandboxes('{}')/tm1.Merge", source_sandbox_name)
        payload = dict()
        payload["Target@odata.bind"] = format_url(
            "Sandboxes('{}')", target_sandbox_name
        )
        payload["CleanAfter"] = clean_after
        return self._rest.POST(url=url, data=json.dumps(payload), **kwargs)

    def exists(self, sandbox_name: str, **kwargs) -> bool:
        """ check if the sandbox exists in TM1

        :param sandbox_name: String
        :return: bool
        """
        url = format_url("/api/v1/Sandboxes('{}')", sandbox_name)
        return self._exists(url, **kwargs)



# -*- coding: utf-8 -*-

import collections
import json
from typing import Dict, List

from TM1py.Objects.TM1Object import TM1Object
from TM1py.Utils import format_url


class ChoreTask(TM1Object):
    """ Abstraction of a Chore Task
    
        A Chore task always conistst of
        - The step integer ID: it's order in the execution plan.
          1 to n, where n is the last Process in the Chore
        - The name of the process to execute
        - The parameters for the process
    
    """

    def __init__(self, step: int, process_name: str, parameters: List[Dict[str, str]]):
        """
        
        :param step: step in the execution order of the Chores' processes. 1 to n, where n the number of processes
        :param process_name: name of the process
        :param parameters: list of dictionaries with 'Name' and 'Value' property:
                            [{
                                'Name': '..',
                                'Value': '..'
                            },
                            ...
                            ]            
        """
        self._step = step
        self._process_name = process_name
        self._parameters = parameters

    @classmethod
    def from_dict(cls, chore_task_as_dict: Dict):
        return cls(step=int(chore_task_as_dict['Step']),
                   process_name=chore_task_as_dict['Process']['Name'],
                   parameters=[{'Name': p['Name'], 'Value': p['Value']} for p in chore_task_as_dict['Parameters']])

    @property
    def body_as_dict(self) -> Dict:
        body_as_dict = collections.OrderedDict()
        body_as_dict['Process@odata.bind'] = format_url("Processes('{}')", self._process_name)
        body_as_dict['Parameters'] = self._parameters
        return body_as_dict

    @property
    def step(self) -> int:
        return self._step

    @property
    def process_name(self) -> str:
        return self._process_name

    @property
    def parameters(self) -> List[Dict[str, str]]:
        return self._parameters

    @property
    def body(self) -> str:
        return json.dumps(self.body_as_dict, ensure_ascii=False)

    def __eq__(self, other: 'ChoreTask') -> bool:
        return self.process_name == other.process_name and self.parameters == other.parameters

    def __ne__(self, other: 'ChoreTask') -> bool:
        return self.process_name != other.process_name or self._parameters != other.parameters



# -*- coding: utf-8 -*-
import json
import warnings
from collections import namedtuple, OrderedDict
from enum import Enum
from typing import Union, Dict

from TM1py.Objects.TM1Object import TM1Object
from TM1py.Utils import format_url

ApplicationType = namedtuple('ApplicationType', ['value', 'suffix', 'odata_type'])


class ApplicationTypes(Enum):
    CHORE = ApplicationType(1, ".chore", "tm1.ChoreReference")
    CUBE = ApplicationType(2, ".cube", "tm1.CubeReference")
    DIMENSION = ApplicationType(3, ".dimension", "tm1.DimensionReference")
    DOCUMENT = ApplicationType(4, ".blob", "#ibm.tm1.api.v1.Document")
    FOLDER = ApplicationType(5, "", "#ibm.tm1.api.v1.Folder")
    LINK = ApplicationType(6, ".extr", "#ibm.tm1.api.v1.Link")
    PROCESS = ApplicationType(7, ".process", "tm1.ProcessReference")
    SUBSET = ApplicationType(8, ".subset", "tm1.SubsetReference")
    VIEW = ApplicationType(9, ".view", "tm1.ViewReference")

    @classmethod
    def _missing_(cls, value: str) -> ApplicationType:
        for member in cls:
            if member.name.lower() == value.lower():
                return member

    @property
    def suffix(self) -> str:
        return self.value.suffix

    @property
    def odata_type(self) -> str:
        return self.value.odata_type


class Application(TM1Object):

    def __init__(self, path: str, name: str, application_type: Union[ApplicationTypes, str]):
        self.path = path
        # remove suffix from name
        if application_type.suffix and name.endswith(application_type.suffix):
            self.name = name[: - len(application_type.suffix)]
        else:
            self.name = name
        # raise ValueError if not a valid type
        self.application_type = ApplicationTypes(application_type)

    @property
    def application_id(self) -> str:
        return self.path + self.name + self.application_type.suffix

    @property
    def body_as_dict(self) -> Dict:
        body_as_dict = OrderedDict()
        body_as_dict["@odata.type"] = self.application_type.odata_type
        body_as_dict["Name"] = self.name
        return body_as_dict

    @property
    def body(self) -> str:
        body_as_dict = self.body_as_dict
        return json.dumps(body_as_dict, ensure_ascii=False)


class ChoreApplication(Application):
    def __init__(self, path: str, name: str, chore_name: str):
        super().__init__(path, name, ApplicationTypes.CHORE)
        self.chore_name = chore_name

    @property
    def body(self) -> str:
        body_as_dict = self.body_as_dict
        body_as_dict["Chore@odata.bind"] = format_url("Chores('{}')", self.chore_name)
        return json.dumps(body_as_dict, ensure_ascii=False)


class CubeApplication(Application):
    def __init__(self, path: str, name: str, cube_name: str):
        super().__init__(path, name, ApplicationTypes.CUBE)
        self.cube_name = cube_name

    @property
    def body(self) -> str:
        body_as_dict = self.body_as_dict
        body_as_dict["Cube@odata.bind"] = format_url("Cubes('{}')", self.cube_name)
        return json.dumps(body_as_dict, ensure_ascii=False)


class DimensionApplication(Application):
    def __init__(self, path: str, name: str, dimension_name: str):
        super().__init__(path, name, ApplicationTypes.DIMENSION)
        self.dimension_name = dimension_name

    @property
    def body(self) -> str:
        body_as_dict = self.body_as_dict
        body_as_dict["Dimension@odata.bind"] = format_url("Dimensions('{}')", self.dimension_name)
        return json.dumps(body_as_dict, ensure_ascii=False)


class DocumentApplication(Application):
    def __init__(self, path: str, name: str, content: bytes, file_id: str = None, file_name: str = None,
                 last_updated: str = None):
        super().__init__(path, name, ApplicationTypes.DOCUMENT)
        self.content = content
        # below fields only populated for retrieved applications
        self.file_id = file_id
        self.file_name = file_name
        self.last_updated = last_updated

    def to_xlsx(self, path_to_file: str):
        warnings.warn(
            f"Function 'to_xlsx' is deprecated. Use 'to_file' instead",
            DeprecationWarning,
            stacklevel=2)
        return self.to_file(path_to_file=path_to_file)

    def to_file(self, path_to_file: str):
        """

        :param path_to_file: path to newly to create file including the extension (e.g., xlsx, xlsm)
        :return:
        """
        with open(path_to_file, "wb") as file:
            file.write(self.content)


class FolderApplication(Application):
    def __init__(self, path: str, name: str):
        super().__init__(path, name, ApplicationTypes.FOLDER)


class LinkApplication(Application):
    def __init__(self, path: str, name: str, url: str):
        super().__init__(path, name, ApplicationTypes.LINK)
        self.url = url

    @property
    def body(self) -> str:
        body_as_dict = self.body_as_dict
        body_as_dict["URL"] = self.url
        return json.dumps(body_as_dict, ensure_ascii=False)


class ProcessApplication(Application):
    def __init__(self, path: str, name: str, process_name: str):
        super().__init__(path, name, ApplicationTypes.PROCESS)
        self.process_name = process_name

    @property
    def body(self) -> str:
        body_as_dict = self.body_as_dict
        body_as_dict["Process@odata.bind"] = format_url("Processes('{}')", self.process_name)
        return json.dumps(body_as_dict, ensure_ascii=False)


class SubsetApplication(Application):
    def __init__(self, path: str, name: str, dimension_name: str, hierarchy_name: str, subset_name: str):
        super().__init__(path, name, ApplicationTypes.SUBSET)
        self.dimension_name = dimension_name
        self.hierarchy_name = hierarchy_name
        self.subset_name = subset_name

    @property
    def body(self) -> str:
        body_as_dict = self.body_as_dict
        body_as_dict["Subset@odata.bind"] = format_url(
            "Dimensions('{}')/Hierarchies('{}')/Subsets('{}')",
            self.dimension_name, self.hierarchy_name, self.subset_name)
        return json.dumps(body_as_dict, ensure_ascii=False)


class ViewApplication(Application):
    def __init__(self, path: str, name: str, cube_name: str, view_name: str):
        super().__init__(path, name, ApplicationTypes.VIEW)
        self.cube_name = cube_name
        self.view_name = view_name

    @property
    def body(self) -> str:
        body_as_dict = self.body_as_dict
        body_as_dict["View@odata.bind"] = format_url("Cubes('{}')/Views('{}')", self.cube_name, self.view_name)
        return json.dumps(body_as_dict, ensure_ascii=False)



# -*- coding: utf-8 -*-
from typing import List

from requests import Response

from TM1py.Objects import Subset
from TM1py.Services.ObjectService import ObjectService
from TM1py.Services.ProcessService import ProcessService
from TM1py.Services.RestService import RestService
from TM1py.Utils import format_url


class SubsetService(ObjectService):
    """ Service to handle Object Updates for TM1 Subsets (dynamic and static)
    
    """

    def __init__(self, rest: RestService):
        super().__init__(rest)
        self._process_service = ProcessService(rest)

    def create(self, subset: Subset, private: bool = False, **kwargs) -> Response:
        """ create subset on the TM1 Server

            :param subset: TM1py.Subset, the subset that shall be created
            :param private: boolean

            :return:
                string: the response
        """
        subsets = "PrivateSubsets" if private else "Subsets"
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/{}",
            subset.dimension_name,
            subset.hierarchy_name,
            subsets)
        response = self._rest.POST(url, subset.body, **kwargs)
        return response

    def get(self, subset_name: str, dimension_name: str, hierarchy_name: str = None, private: bool = False,
            **kwargs) -> Subset:
        """ get a subset from the TM1 Server

            :param subset_name: string, name of the subset
            :param dimension_name: string, name of the dimension
            :param hierarchy_name: string, name of the hierarchy
            :param private: Boolean

            :return: instance of TM1py.Subset
        """
        if not hierarchy_name:
            hierarchy_name = dimension_name
        subsets = "PrivateSubsets" if private else "Subsets"
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/{}('{}')?$expand=Hierarchy($select=Dimension,Name),"
            "Elements($select=Name)&$select=*,Alias", dimension_name, hierarchy_name, subsets, subset_name)
        response = self._rest.GET(url=url, **kwargs)
        return Subset.from_dict(response.json())

    def get_all_names(self, dimension_name: str, hierarchy_name: str = None, private: bool = False,
                      **kwargs) -> List[str]:
        """ get names of all private or public subsets in a hierarchy

        :param dimension_name:
        :param hierarchy_name:
        :param private: Boolean
        :return: List of Strings
        """
        hierarchy_name = hierarchy_name if hierarchy_name else dimension_name

        subsets = "PrivateSubsets" if private else "Subsets"
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/{}?$select=Name",
            dimension_name, hierarchy_name, subsets)
        response = self._rest.GET(url=url, **kwargs)
        subsets = response.json()['value']
        return [subset['Name'] for subset in subsets]

    def update(self, subset: Subset, private: bool = False, **kwargs) -> Response:
        """ update a subset on the TM1 Server

        :param subset: instance of TM1py.Subset.
        :param private: Boolean
        :return: response
        """
        if subset.is_static:
            self.delete_elements_from_static_subset(
                dimension_name=subset.dimension_name,
                hierarchy_name=subset.hierarchy_name,
                subset_name=subset.name,
                private=private,
                **kwargs)
        subsets = "PrivateSubsets" if private else "Subsets"
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/{}('{}')",
            subset.dimension_name, subset.hierarchy_name, subsets, subset.name)
        return self._rest.PATCH(url=url, data=subset.body, **kwargs)

    def make_static(self, subset_name: str, dimension_name: str, hierarchy_name: str = None,
                    private: bool = False) -> Response:
        """ convert a dynamic subset into static subset on the TM1 Server
        :param subset_name: String, name of the subset
        :param dimension_name: String, name of the dimension
        :param hierarchy_name: String, name of the hierarchy
        :param private: Boolean
        :return: response
        """
        import json
        from collections import OrderedDict
        hierarchy_name = hierarchy_name if hierarchy_name else dimension_name
        payload = OrderedDict()
        payload['Name'] = subset_name
        payload['MakePrivate'] = True if private else False
        payload['MakeStatic'] = True
        subsets = "PrivateSubsets" if private else "Subsets"
        url = format_url("/api/v1/Dimensions('{}')/Hierarchies('{}')/{}('{}')/tm1.SaveAs", dimension_name,
                         hierarchy_name, subsets, subset_name)
        return self._rest.POST(url=url, data=json.dumps(payload))

    def update_or_create(self, subset: Subset, private: bool = False, **kwargs) -> Response:
        """ update if exists else create

        :param subset:
        :param private:
        :return:
        """
        if self.exists(
                subset_name=subset.name,
                dimension_name=subset.dimension_name,
                hierarchy_name=subset.hierarchy_name,
                private=private,
                **kwargs):
            return self.update(subset=subset, private=private, **kwargs)

        return self.create(subset=subset, private=private, **kwargs)

    def delete(self, subset_name: str, dimension_name: str, hierarchy_name: str = None,
               private: bool = False, **kwargs) -> Response:
        """ Delete an existing subset on the TM1 Server

        :param subset_name: String, name of the subset
        :param dimension_name: String, name of the dimension
        :param hierarchy_name: String, name of the hierarchy
        :param private: Boolean
        :return:
        """
        hierarchy_name = hierarchy_name if hierarchy_name else dimension_name
        subsets = "PrivateSubsets" if private else "Subsets"
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/{}('{}')",
            dimension_name, hierarchy_name, subsets, subset_name)
        response = self._rest.DELETE(url=url, **kwargs)
        return response

    def exists(self, subset_name: str, dimension_name: str, hierarchy_name: str = None, private: bool = False,
               **kwargs) -> bool:
        """checks if private or public subset exists

        :param subset_name: 
        :param dimension_name: 
        :param hierarchy_name:
        :param private:
        :return: boolean
        """
        hierarchy_name = hierarchy_name if hierarchy_name else dimension_name
        subset_type = 'PrivateSubsets' if private else "Subsets"
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/{}('{}')",
            dimension_name, hierarchy_name, subset_type, subset_name)
        return self._exists(url, **kwargs)

    def delete_elements_from_static_subset(self, dimension_name: str, hierarchy_name: str, subset_name: str,
                                           private: bool, **kwargs) -> Response:
        subsets = "PrivateSubsets" if private else "Subsets"
        url = format_url(
            "/api/v1/Dimensions('{}')/Hierarchies('{}')/{}('{}')/Elements/$ref",
            dimension_name, hierarchy_name, subsets, subset_name)
        return self._rest.DELETE(url=url, **kwargs)

    def get_element_names(self, dimension_name: str, hierarchy_name: str, subset_name: str, private: bool = False,
                          **kwargs):
        """ Get elements from existing (dynamic or static) subset

        :param dimension_name:
        :param hierarchy_name:
        :param subset_name:
        :param private:
        :param kwargs:
        :return:
        """
        subset = self.get(subset_name, dimension_name, hierarchy_name, private=private, **kwargs)
        if subset.is_static:
            return subset.elements

        mdx = subset.expression
        from TM1py import ElementService
        element_service = ElementService(self._rest)
        tuples = element_service.execute_set_mdx(
            mdx=mdx,
            member_properties=["Name"],
            element_properties=None,
            parent_properties=None,
            **kwargs)
        return [entry[0]["Name"] for entry in tuples]



# -*- coding: utf-8 -*-

import functools
import json
from collections.abc import Iterable
from datetime import datetime
from typing import Dict, Optional

import pytz
from requests import Response
from TM1py.Objects.Process import Process

from TM1py.Services.ObjectService import ObjectService
from TM1py.Services.RestService import RestService
from TM1py.Utils import format_url
from TM1py.Utils.Utils import CaseAndSpaceInsensitiveDict, CaseAndSpaceInsensitiveSet, require_admin, require_version


def odata_track_changes_header(func):
    """ Higher Order function to handle addition and removal of odata.track-changes HTTP Header

    :param func: 
    :return: 
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        # Add header
        self._rest.add_http_header("Prefer", "odata.track-changes")
        # Do stuff
        response = func(self, *args, **kwargs)
        # Remove Header
        self._rest.remove_http_header("Prefer")
        return response

    return wrapper


class ServerService(ObjectService):
    """ Service to query common information from the TM1 Server
    
    """

    def __init__(self, rest: RestService):
        super().__init__(rest)
        self.tlog_last_delta_request = None
        self.mlog_last_delta_request = None
        self.alog_last_delta_request = None

    @odata_track_changes_header
    def initialize_transaction_log_delta_requests(self, filter=None, **kwargs):
        url = "/api/v1/TailTransactionLog()"
        if filter:
            url += "?$filter={}".format(filter)
        response = self._rest.GET(url=url, **kwargs)
        # Read the next delta-request-url from the response
        self.tlog_last_delta_request = response.text[response.text.rfind("TransactionLogEntries/!delta('"):-2]

    @odata_track_changes_header
    def execute_transaction_log_delta_request(self, **kwargs) -> Dict:
        response = self._rest.GET(url="/api/v1/" + self.tlog_last_delta_request, **kwargs)
        self.tlog_last_delta_request = response.text[response.text.rfind("TransactionLogEntries/!delta('"):-2]
        return response.json()['value']

    @odata_track_changes_header
    def initialize_audit_log_delta_requests(self, filter=None, **kwargs):
        url = "/api/v1/TailAuditLog()"
        if filter:
            url += "?$filter={}".format(filter)
        response = self._rest.GET(url=url, **kwargs)
        # Read the next delta-request-url from the response
        self.alog_last_delta_request = response.text[response.text.rfind("AuditLogEntries/!delta('"):-2]

    @odata_track_changes_header
    def execute_audit_log_delta_request(self, **kwargs) -> Dict:
        response = self._rest.GET(url="/api/v1/" + self.alog_last_delta_request, **kwargs)
        self.alog_last_delta_request = response.text[response.text.rfind("AuditLogEntries/!delta('"):-2]
        return response.json()['value']

    @odata_track_changes_header
    def initialize_message_log_delta_requests(self, filter=None, **kwargs):
        url = "/api/v1/TailMessageLog()"
        if filter:
            url += "?$filter={}".format(filter)
        response = self._rest.GET(url=url, **kwargs)
        # Read the next delta-request-url from the response
        self.mlog_last_delta_request = response.text[response.text.rfind("MessageLogEntries/!delta('"):-2]

    @odata_track_changes_header
    def execute_message_log_delta_request(self, **kwargs) -> Dict:
        response = self._rest.GET(url="/api/v1/" + self.mlog_last_delta_request, **kwargs)
        self.mlog_last_delta_request = response.text[response.text.rfind("MessageLogEntries/!delta('"):-2]
        return response.json()['value']

    @require_admin
    def get_message_log_entries(self, reverse: bool = True, since: datetime = None,
                                until: datetime = None, top: int = None, logger: str = None,
                                level: str = None, msg_contains: Iterable = None, msg_contains_operator: str = 'and',
                                **kwargs) -> Dict:
        """
        :param reverse: Boolean
        :param since: of type datetime. If it doesn't have tz information, UTC is assumed.
        :param until: of type datetime. If it doesn't have tz information, UTC is assumed.
        :param top: Integer
        :param logger: string, eg TM1.Server, TM1.Chore, TM1.Mdx.Interface, TM1.Process
        :param level: string, ERROR, WARNING, INFO, DEBUG, UNKNOWN
        :param msg_contains: iterable, find substring in log message; list of substrings will be queried as AND statement
        :param msg_contains_operator: 'and' or 'or'

        :param kwargs:
        :return: Dict of server log
        """
        msg_contains_operator = msg_contains_operator.strip().lower()
        if msg_contains_operator not in ("and", "or"):
            raise ValueError("'msg_contains_operator' must be either 'AND' or 'OR'")

        reverse = 'desc' if reverse else 'asc'
        url = '/api/v1/MessageLogEntries?$orderby=TimeStamp {}'.format(reverse)

        if since or until or logger or level or msg_contains:
            log_filters = []

            if since:
                # If since doesn't have tz information, UTC is assumed
                if not since.tzinfo:
                    since = self.utc_localize_time(since)
                log_filters.append(format_url("TimeStamp ge {}", since.strftime("%Y-%m-%dT%H:%M:%SZ")))

            if until:
                # If until doesn't have tz information, UTC is assumed
                if not until.tzinfo:
                    until = self.utc_localize_time(until)
                log_filters.append(format_url("TimeStamp le {}", until.strftime("%Y-%m-%dT%H:%M:%SZ")))

            if logger:
                log_filters.append(format_url("Logger eq '{}'", logger))

            if level:
                level_dict = CaseAndSpaceInsensitiveDict(
                    {'ERROR': 1, 'WARNING': 2, 'INFO': 3, 'DEBUG': 4, 'UNKNOWN': 5})
                level_index = level_dict.get(level)
                if level_index:
                    log_filters.append("Level eq {}".format(level_index))

            if msg_contains:
                if isinstance(msg_contains, str):
                    log_filters.append(format_url("contains(toupper(Message),toupper('{}'))", msg_contains))
                else:
                    msg_filters = [format_url("contains(toupper(Message),toupper('{}'))", wildcard)
                                   for wildcard in msg_contains]
                    log_filters.append("({})".format(f" {msg_contains_operator} ".join(msg_filters)))

            url += "&$filter={}".format(" and ".join(log_filters))

        if top:
            url += '&$top={}'.format(top)

        response = self._rest.GET(url, **kwargs)
        return response.json()['value']

    @require_admin
    def write_to_message_log(self, level: str, message: str, **kwargs) -> None:
        """
        :param level: string, FATAL, ERROR, WARN, INFO, DEBUG
        :param message: string
        :return:
        """    

        valid_levels = CaseAndSpaceInsensitiveSet({'FATAL', 'ERROR', 'WARN', 'INFO', 'DEBUG'})
        if level not in valid_levels:
            raise ValueError(f"Invalid level: '{level}'")

        from TM1py.Services import ProcessService
        process_service = ProcessService(self._rest)
        process = Process(name="", prolog_procedure="LogOutput('{}', '{}');".format(level, message))
        success, status, _ = process_service.execute_process_with_return(process, **kwargs)

        if not success:
            raise RuntimeError(f"Failed to write to TM1 Message Log through unbound process. Status: '{status}'")

    @staticmethod
    def utc_localize_time(timestamp):
        timestamp = pytz.utc.localize(timestamp)
        timestamp_utc = timestamp.astimezone(pytz.utc)
        return timestamp_utc

    @require_admin
    def get_transaction_log_entries(self, reverse: bool = True, user: str = None, cube: str = None,
                                    since: datetime = None, until: datetime = None, top: int = None, **kwargs) -> Dict:
        """
        :param reverse: Boolean
        :param user: UserName
        :param cube: CubeName
        :param since: of type datetime. If it doesn't have tz information, UTC is assumed.
        :param until: of type datetime. If it doesn't have tz information, UTC is assumed.
        :param top: int
        :return:
        """
        reverse = 'desc' if reverse else 'asc'
        url = '/api/v1/TransactionLogEntries?$orderby=TimeStamp {} '.format(reverse)
        # filter on user, cube and time
        if user or cube or since or until:
            log_filters = []
            if user:
                log_filters.append(format_url("User eq '{}'", user))
            if cube:
                log_filters.append(format_url("Cube eq '{}'", cube))
            if since:
                # If since doesn't have tz information, UTC is assumed
                if not since.tzinfo:
                    since = self.utc_localize_time(since)
                log_filters.append(format_url("TimeStamp ge {}", since.strftime("%Y-%m-%dT%H:%M:%SZ")))
            if until:
                # If until doesn't have tz information, UTC is assumed
                if not until.tzinfo:
                    until = self.utc_localize_time(until)
                log_filters.append(format_url("TimeStamp le {}", until.strftime("%Y-%m-%dT%H:%M:%SZ")))
            url += "&$filter={}".format(" and ".join(log_filters))
        # top limit
        if top:
            url += '&$top={}'.format(top)
        response = self._rest.GET(url, **kwargs)
        return response.json()['value']

    @require_admin
    @require_version(version="11.6")
    def get_audit_log_entries(self, user: str = None, object_type: str = None, object_name: str = None,
                              since: datetime = None, until: datetime = None, top: int = None, **kwargs) -> Dict:
        """
        :param user: UserName
        :param object_type: ObjectType
        :param object_name: ObjectName
        :param since: of type datetime. If it doesn't have tz information, UTC is assumed.
        :param until: of type datetime. If it doesn't have tz information, UTC is assumed.
        :param top: int
        :return:
        """

        url = '/api/v1/AuditLogEntries?$expand=AuditDetails'
        # filter on user, object_type, object_name  and time
        if any([user, object_type, object_name, since, until]):
            log_filters = []
            if user:
                log_filters.append(format_url("UserName eq '{}'", user))
            if object_type:
                log_filters.append(format_url("ObjectType eq '{}'", object_type))
            if object_name:
                log_filters.append(format_url("ObjectName eq '{}'", object_name))
            if since:
                # If since doesn't have tz information, UTC is assumed
                if not since.tzinfo:
                    since = self.utc_localize_time(since)
                log_filters.append(format_url("TimeStamp ge {}", since.strftime("%Y-%m-%dT%H:%M:%SZ")))
            if until:
                # If until doesn't have tz information, UTC is assumed
                if not until.tzinfo:
                    until = self.utc_localize_time(until)
                log_filters.append(format_url("TimeStamp le {}", until.strftime("%Y-%m-%dT%H:%M:%SZ")))
            url += "&$filter={}".format(" and ".join(log_filters))
        # top limit
        if top:
            url += '&$top={}'.format(top)
        response = self._rest.GET(url, **kwargs)
        return response.json()['value']

    @require_admin
    def get_last_process_message_from_messagelog(self, process_name: str, **kwargs) -> Optional[str]:
        """ Get the latest message log entry for a process

            :param process_name: name of the process
            :return: String - the message, for instance: "Ausführung normal beendet, verstrichene Zeit 0.03  Sekunden"
        """
        url = format_url(
            "/api/v1/MessageLog()?$orderby='TimeStamp'&$filter=Logger eq 'TM1.Process' and contains(Message, '{}')",
            process_name)
        response = self._rest.GET(url=url, **kwargs)
        response_as_list = response.json()['value']
        if len(response_as_list) > 0:
            message_log_entry = response_as_list[0]
            return message_log_entry['Message']

    def get_server_name(self, **kwargs) -> str:
        """ Ask TM1 Server for its name

        :Returns:
            String, the server name
        """
        url = '/api/v1/Configuration/ServerName/$value'
        return self._rest.GET(url, **kwargs).text

    def get_product_version(self, **kwargs) -> str:
        """ Ask TM1 Server for its version

        :Returns:
            String, the version
        """
        url = '/api/v1/Configuration/ProductVersion/$value'
        return self._rest.GET(url, **kwargs).text

    def get_admin_host(self, **kwargs) -> str:
        url = '/api/v1/Configuration/AdminHost/$value'
        return self._rest.GET(url, **kwargs).text

    def get_data_directory(self, **kwargs) -> str:
        url = '/api/v1/Configuration/DataBaseDirectory/$value'
        return self._rest.GET(url, **kwargs).text

    def get_configuration(self, **kwargs) -> Dict:
        url = '/api/v1/Configuration'
        config = self._rest.GET(url, **kwargs).json()
        del config["@odata.context"]
        return config

    @require_admin
    def get_static_configuration(self, **kwargs) -> Dict:
        """ Read TM1 config settings as dictionary from TM1 Server

        :return: config as dictionary
        """
        url = '/api/v1/StaticConfiguration'
        config = self._rest.GET(url, **kwargs).json()
        del config["@odata.context"]
        return config

    @require_admin
    def get_active_configuration(self, **kwargs) -> Dict:
        """ Read effective(!) TM1 config settings as dictionary from TM1 Server

        :return: config as dictionary
        """
        url = '/api/v1/ActiveConfiguration'
        config = self._rest.GET(url, **kwargs).json()
        del config["@odata.context"]
        return config

    @require_admin
    def update_static_configuration(self, configuration: Dict) -> Response:
        """ Update the .cfg file and triggers TM1 to re-read the file.

        :param configuration:
        :return: Response
        """
        url = '/api/v1/StaticConfiguration'
        return self._rest.PATCH(url, json.dumps(configuration))

    @require_admin
    def save_data(self, **kwargs) -> Response:
        from TM1py.Services import ProcessService
        ti = "SaveDataAll;"
        process_service = ProcessService(self._rest)
        return process_service.execute_ti_code(ti, **kwargs)

    @require_admin
    def delete_persistent_feeders(self, **kwargs) -> Response:
        from TM1py.Services import ProcessService
        ti = "DeleteAllPersistentFeeders;"
        process_service = ProcessService(self._rest)
        return process_service.execute_ti_code(ti, **kwargs)

    @require_admin
    def start_performance_monitor(self):
        config = {
            "Administration": {"PerformanceMonitorOn": True}
        }
        self.update_static_configuration(config)

    @require_admin
    def stop_performance_monitor(self):
        config = {
            "Administration": {"PerformanceMonitorOn": False}
        }
        self.update_static_configuration(config)

    @require_admin
    def activate_audit_log(self):
        config = {'Administration': {'AuditLog': {'Enable': True}}}
        self.update_static_configuration(config)

    @require_admin
    def deactivate_audit_log(self):
        config = {'Administration': {'AuditLog': {'Enable': False}}}
        self.update_static_configuration(config)



import warnings


class DimensionSelection:
    """ Instances of this class to be passed to construct_mdx function

    """
    SUBSET = 1
    EXPRESSION = 2
    ITERABLE = 3

    def __init__(self, dimension_name, elements=None, subset=None, expression=None):
        warnings.warn(
            f"class DimensionSelection will be deprecated. Use https://github.com/cubewise-code/mdxpy instead",
            DeprecationWarning,
            stacklevel=2)
        self.dimension_name = dimension_name
        self.selection_type = self.determine_selection_type(elements, subset, expression)
        if self.selection_type == self.SUBSET:
            self.expression = curly_braces(expression="Tm1SubsetToSet([{dimension}], '{subset}')".format(
                dimension=dimension_name,
                subset=subset))
        elif self.selection_type == self.EXPRESSION:
            self.expression = curly_braces(expression=expression)
        elif self.selection_type == self.ITERABLE:
            self.expression = curly_braces(expression=",".join(["[{}].[{}]".format(dimension_name, element)
                                                                for element
                                                                in elements]))
        elif not self.selection_type:
            self.expression = curly_braces(expression="TM1SubsetAll([{dimension}])".format(dimension=dimension_name))

    @staticmethod
    def determine_selection_type(elements=None, subset=None, expression=None):
        warnings.warn(
            f"Module MdxUtils will be deprecated. Use https://github.com/cubewise-code/mdxpy instead",
            DeprecationWarning,
            stacklevel=2)
        if elements is not None and subset is None and expression is None:
            return DimensionSelection.ITERABLE
        elif elements is None and subset is not None and expression is None:
            return DimensionSelection.SUBSET
        elif elements is None and subset is None and expression is not None:
            return DimensionSelection.EXPRESSION
        elif elements is None and subset is None and expression is None:
            return None
        else:
            raise ValueError("DimensionSelection constructor takes one type of selection only: "
                             "elements, subset or expression")


def construct_mdx_axis(dim_selections):
    """ Construct MDX for one Axis (Row or Column).
    Can have multiple dimensions stacked.

    :param dim_selections: instances of TM1py.Utils.MDXUtils.DimensionSelection
    :return: a valid MDX for an Axis
    """
    warnings.warn(
        f"Module MdxUtils will be deprecated. Use https://github.com/cubewise-code/mdxpy instead",
        DeprecationWarning,
        stacklevel=2)
    return "*".join(selection.expression
                    for selection
                    in dim_selections)


def construct_mdx(cube_name, rows, columns, contexts=None, suppress=None):
    """ Method to construct MDX Query from different dimension selection

    :param cube_name: Name of the Cube
    :param rows: List of DimensionSelections
    :param columns: List of DimensionSelections
    :param contexts: Dictionary of Dimensions and Elements
    :param suppress: "Both", "Rows", "Columns" or None
    :return: Generated MDX Query
    """
    warnings.warn(
        f"Module MdxUtils will be deprecated. Use https://github.com/cubewise-code/mdxpy instead",
        DeprecationWarning,
        stacklevel=2)
    # MDX Skeleton
    mdx_template = "SELECT {}{} ON ROWS, {}{} ON COLUMNS FROM [{}] {}"
    # Suppression
    mdx_rows_suppress = "NON EMPTY " if suppress and suppress.upper() in ["ROWS", "BOTH"] else ""
    mdx_columns_suppress = "NON EMPTY " if suppress and suppress.upper() in ["COLUMNS", "BOTH"] else ""
    # Rows and Columns
    mdx_rows = construct_mdx_axis(rows)
    mdx_columns = construct_mdx_axis(columns)
    # Context filter (where statement)
    mdx_where = ""
    if contexts:
        mdx_where_parts = ["[{}].[{}]".format(dim, elem)
                           for dim, elem
                           in contexts.items()]
        mdx_where = "".join(["WHERE (",
                             ",".join(mdx_where_parts),
                             ")"])
    # Return Full MDX
    return mdx_template.format(mdx_rows_suppress, mdx_rows, mdx_columns_suppress, mdx_columns, cube_name, mdx_where)


def curly_braces(expression):
    """ Put curly braces around a string

    :param expression:
    :return:
    """
    warnings.warn(
        f"Module MdxUtils will be deprecated. Use https://github.com/cubewise-code/mdxpy instead",
        DeprecationWarning,
        stacklevel=2)
    return "".join(["{" if not expression.startswith("{") else "",
                    expression,
                    "}" if not expression.endswith("}") else ""])


def read_cube_name_from_mdx(mdx):
    """ Read the cube name from a valid MDX Query

    :param mdx: The MDX Query as String
    :return: String, name of a cube
    """
    warnings.warn(
        f"Module MdxUtils will be deprecated. Use https://github.com/cubewise-code/mdxpy instead",
        DeprecationWarning,
        stacklevel=2)
    cube, _, _, _ = read_dimension_composition_from_mdx(mdx)
    return cube


def read_dimension_composition_from_mdx(mdx):
    """ Parse a valid MDX Query and return the name of the cube and a list of dimensions for each axis

    :param mdx:
    :return:
    """
    warnings.warn(
        f"Module MdxUtils will be deprecated. Use https://github.com/cubewise-code/mdxpy instead",
        DeprecationWarning,
        stacklevel=2)
    mdx_rows, mdx_columns, mdx_from, mdx_where = split_mdx(mdx)

    cube = mdx_from[1:-1]
    rows = read_dimension_composition_from_mdx_set_or_tuple(mdx_rows)
    columns = read_dimension_composition_from_mdx_set_or_tuple(mdx_columns)
    titles = read_dimension_composition_from_mdx_set_or_tuple(mdx_where)

    return cube, rows, columns, titles


def read_dimension_composition_from_mdx_set_or_tuple(mdx):
    warnings.warn(
        f"Module MdxUtils will be deprecated. Use https://github.com/cubewise-code/mdxpy instead",
        DeprecationWarning,
        stacklevel=2)
    mdx_without_spaces = ''.join(mdx.split())
    # case for mdx statement no where statement
    if len(mdx_without_spaces) == 0:
        return []
    # case for tuples mdx statement on rows or columns
    if mdx_without_spaces[1] == '(' and mdx_without_spaces[-2] == ')':
        return read_dimension_composition_from_mdx_tuple(mdx)
    # case for where mdx statement
    elif mdx_without_spaces[0] == '(' and mdx_without_spaces[-1] == ')':
        return read_dimension_composition_from_mdx_tuple(mdx)
    # case for set mdx statement on rows or columns
    else:
        return read_dimension_composition_from_mdx_set(mdx)


def read_dimension_composition_from_mdx_set(mdx):
    warnings.warn(
        f"Module MdxUtils will be deprecated. Use https://github.com/cubewise-code/mdxpy instead",
        DeprecationWarning,
        stacklevel=2)
    dimensions = []
    mdx_without_spaces = ''.join(mdx.split())
    for sub_mdx in mdx_without_spaces.split("}*{"):
        pos_start, pos_end = sub_mdx.find("["), sub_mdx.find("]")
        dimension_name = sub_mdx[pos_start + 1:pos_end]
        dimensions.append(dimension_name)
    return dimensions


def read_dimension_composition_from_mdx_tuple(mdx):
    warnings.warn(
        f"Module MdxUtils will be deprecated. Use https://github.com/cubewise-code/mdxpy instead",
        DeprecationWarning,
        stacklevel=2)
    dimensions = []
    for unique_member_name in mdx.split(","):
        pos_start, pos_end = unique_member_name.find("["), unique_member_name.find("]")
        dimension_name = unique_member_name[pos_start + 1:pos_end]
        # only parse through first tuple of potentially many tuples
        if dimension_name in dimensions:
            return dimensions
        dimensions.append(dimension_name)
    return dimensions


def split_mdx(mdx):
    warnings.warn(
        f"Module MdxUtils will be deprecated. Use https://github.com/cubewise-code/mdxpy instead",
        DeprecationWarning,
        stacklevel=2)
    try:
        mdx_rows, mdx_rest = _find_case_and_space_insensitive_first_occurrence(
            text=mdx,
            pattern_start="{",
            pattern_end="}ONROWS"
        )
        mdx_columns, mdx_rest = _find_case_and_space_insensitive_first_occurrence(
            text=mdx_rest,
            pattern_start="{",
            pattern_end="}ONCOLUMNSFROM"
        )
        mdx_from, mdx_where = _find_case_and_space_insensitive_first_occurrence(
            text=mdx_rest,
            pattern_end="]WHERE"
        )
        return mdx_rows, mdx_columns, mdx_from, mdx_where
    except ValueError:
        ValueError("Can't parse mdx: {}".format(mdx))


def _find_case_and_space_insensitive_first_occurrence(text, pattern_start=None, pattern_end=None):
    warnings.warn(
        f"Module MdxUtils will be deprecated. Use https://github.com/cubewise-code/mdxpy instead",
        DeprecationWarning,
        stacklevel=2)
    text_without_spaces = ''.join(text.split())
    text_without_spaces_and_uppercase = text_without_spaces.upper()

    if pattern_start:
        pattern_start = ''.join(pattern_start.split()).upper()
    if pattern_end:
        pattern_end = ''.join(pattern_end.split()).upper()

    if text_without_spaces_and_uppercase.count(pattern_end) > 1:
        raise ValueError("Invalid state. {} has more than 1 occurrences in text: {}".format(pattern_end, text))
    pos_start = text_without_spaces_and_uppercase.find(pattern_start) if pattern_start else 0
    pos_end = text_without_spaces_and_uppercase.find(pattern_end) if pattern_end else -1

    # case of mdx statement without where clause
    if pos_start == 0 and pos_end == -1:
        return text, ""
    selection = text_without_spaces[pos_start:pos_end + 1]
    text = text_without_spaces[pos_end + len(pattern_end):]
    return selection, text



# -*- coding: utf-8 -*-

import functools
import json
from datetime import datetime
from typing import List

from requests import Response

from TM1py.Objects import Chore, ChoreTask
from TM1py.Services.ObjectService import ObjectService
from TM1py.Services.RestService import RestService
from TM1py.Utils import format_url


def deactivate_activate(func):
    """ Higher Order function to handle activation and deactivation of chores before updating them

    :param func:
    :return:
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        # chore (type Chore) or chore_name (type str) is passed as first arg or kwargs
        if 'chore' in kwargs:
            chore = kwargs.get('chore').name
        elif 'chore_name' in kwargs:
            chore = kwargs.get('chore_name')
        else:
            chore = args[0]

        reactivate = None
        if isinstance(chore, Chore):
            chore_name = chore.name
            reactivate = chore.active
        elif isinstance(chore, str):
            chore_name = chore
        else:
            raise ValueError("Argument must be of type 'Chore' or 'str'")

        # Get current chore
        chore_old = self.get(chore_name)
        if reactivate is None:
            reactivate = chore_old.active

        # Deactivate
        if chore_old.active:
            self.deactivate(chore_name)
        # Do stuff
        try:
            response = func(self, *args, **kwargs)
        except Exception as e:
            raise e
        # Activate if necessary
        finally:
            if reactivate:
                self.activate(chore_name)
        return response

    return wrapper


class ChoreService(ObjectService):
    """ Service to handle Object Updates for TM1 Chores

    """

    def __init__(self, rest: RestService):
        super().__init__(rest)

    def get(self, chore_name: str, **kwargs) -> Chore:
        """ Get a chore from the TM1 Server
        :param chore_name:
        :return: instance of TM1py.Chore
        """
        url = format_url(
            "/api/v1/Chores('{}')?$expand=Tasks($expand=*,Process($select=Name),Chore($select=Name))",
            chore_name)
        response = self._rest.GET(url, **kwargs)
        return Chore.from_dict(response.json())

    def get_all(self, **kwargs) -> List[Chore]:
        """ get a List of all Chores
        :return: List of TM1py.Chore
        """
        url = "/api/v1/Chores?$expand=Tasks($expand=*,Process($select=Name),Chore($select=Name))"
        response = self._rest.GET(url, **kwargs)
        return [Chore.from_dict(chore_as_dict) for chore_as_dict in response.json()['value']]

    def get_all_names(self, **kwargs) -> List[str]:
        """ get a List of all Chores
        :return: List of TM1py.Chore
        """
        url = "/api/v1/Chores?$select=Name"
        response = self._rest.GET(url, **kwargs)
        return [chore['Name'] for chore in response.json()['value']]

    def create(self, chore: Chore, **kwargs) -> Response:
        """ create a chore
        :param chore: instance of TM1py.Chore
        :return:
        """
        url = "/api/v1/Chores"
        response = self._rest.POST(url=url, data=chore.body, **kwargs)

        if chore.dst_sensitivity:
            self.set_local_start_time(chore.name, chore.start_time.datetime)

        if chore.active:
            self.activate(chore.name)
        return response

    def delete(self, chore_name: str, **kwargs) -> Response:
        """ delete chore in TM1
        :param chore_name:
        :return: response
        """
        url = format_url("/api/v1/Chores('{}')", chore_name)
        response = self._rest.DELETE(url, **kwargs)
        return response

    def exists(self, chore_name: str, **kwargs) -> bool:
        """ Check if Chore exists

        :param chore_name:
        :return:
        """
        url = format_url("/api/v1/Chores('{}')", chore_name)
        return self._exists(url, **kwargs)

    def search_for_process_name(self, process_name: str, **kwargs) -> List[Chore]:
        """ Return chore details for any/all chores that contain specified process name in tasks

        :param process_name: string, a valid ti process name; spaces will be elimniated
        """
        url = format_url(
            "/api/v1/Chores?$filter=Tasks/any(t: replace(tolower(t/Process/Name), ' ', '') eq '{}')"
            "&$expand=Tasks($expand=*,Chore($select=Name),Process($select=Name))",
            process_name.lower().replace(' ', '')
        )
        response = self._rest.GET(url, **kwargs)
        return [Chore.from_dict(chore_as_dict) for chore_as_dict in response.json()['value']]

    def search_for_parameter_value(self, parameter_value: str, **kwargs) -> List[Chore]:
        """ Return chore details for any/all chores that have a specified value set in the chore parameter settings
            *this will NOT check the process parameter default, rather the defined parameter value saved in the chore

        :param parameter_value: string, will search wildcard for string in parameter value using Contains(string)
        """
        url = format_url(
            "/api/v1/Chores?"
            "$filter=Tasks/any(t: t/Parameters/any(p: isof(p/Value, Edm.String) and contains(tolower(p/Value), '{}')))"
            "&$expand=Tasks($expand=*,Process($select=Name),Chore($select=Name))",
            parameter_value.lower()
        )
        response = self._rest.GET(url, **kwargs)
        return [Chore.from_dict(chore_as_dict) for chore_as_dict in response.json()['value']]

    @deactivate_activate
    def update(self, chore: Chore, **kwargs):
        """ update chore on TM1 Server
        does not update: DST Sensitivity!
        :param chore:
        :return:
        """
        # Update StartTime, ExecutionMode, Frequency
        url = format_url("/api/v1/Chores('{}')", chore.name)
        # Remove Tasks from Body. Tasks to be managed individually
        chore_dict_without_tasks = chore.body_as_dict
        chore_dict_without_tasks.pop("Tasks")
        self._rest.PATCH(url, json.dumps(chore_dict_without_tasks), **kwargs)

        # Update Tasks individually
        task_old_count = self._get_tasks_count(chore.name)
        for i, task_new in enumerate(chore.tasks):
            if i >= task_old_count:
                self._add_task(chore.name, task_new, **kwargs)
            else:
                task_old = self._get_task(chore.name, i)
                if task_new != task_old:
                    self._update_task(chore.name, task_new, **kwargs)
        for j in range(i + 1, task_old_count):
            self._delete_task(chore.name, i + 1, **kwargs)

        if chore.dst_sensitivity:
            self.set_local_start_time(chore.name, chore.start_time.datetime)

    def update_or_create(self, chore: Chore, **kwargs) -> Response:
        if self.exists(chore_name=chore.name, **kwargs):
            return self.update(chore, **kwargs)

        return self.create(chore=chore, **kwargs)

    def activate(self, chore_name: str, **kwargs) -> Response:
        """ activate chore on TM1 Server
        :param chore_name:
        :return: response
        """
        url = format_url("/api/v1/Chores('{}')/tm1.Activate", chore_name)
        return self._rest.POST(url, '', **kwargs)

    def deactivate(self, chore_name: str, **kwargs) -> Response:
        """ deactivate chore on TM1 Server
        :param chore_name:
        :return: response
        """
        url = format_url("/api/v1/Chores('{}')/tm1.Deactivate", chore_name)
        return self._rest.POST(url, '', **kwargs)

    @deactivate_activate
    def set_local_start_time(self, chore_name: str, date_time: datetime, **kwargs) -> Response:
        """ Makes Server crash if chore is activated (10.2.2 FP6) :)
        :param chore_name:
        :param date_time:
        :return:
        """
        url = format_url("/api/v1/Chores('{}')/tm1.SetServerLocalStartTime", chore_name)
        data = {
            "StartDate": "{}-{}-{}".format(
                date_time.year, date_time.month, date_time.day),
            "StartTime": "{}:{}:{}".format(
                self.zfill_two(date_time.hour), self.zfill_two(date_time.minute), self.zfill_two(date_time.second))
        }
        return self._rest.POST(url, json.dumps(data), **kwargs)

    def execute_chore(self, chore_name: str, **kwargs) -> Response:
        """ Ask TM1 Server to execute a chore
            :param chore_name: String, name of the chore to be executed
            :return: the response
        """
        return self._rest.POST(format_url("/api/v1/Chores('{}')/tm1.Execute", chore_name), '', **kwargs)

    def _get_tasks_count(self, chore_name: str, **kwargs) -> int:
        """ Query Chore tasks count on TM1 Server
        :param chore_name: name of Chore to count tasks
        :return: int
        """
        url = format_url("/api/v1/Chores('{}')/Tasks/$count", chore_name)
        response = self._rest.GET(url, **kwargs)
        return int(response.text)

    def _get_task(self, chore_name: str, step: int, **kwargs) -> ChoreTask:
        """ Get task from chore
        :param chore_name: name of the chore
        :param step:
        :return: instance of TM1py.ChoreTask
        """
        url = format_url(
            "/api/v1/Chores('{}')/Tasks({})?$expand=*,Process($select=Name),Chore($select=Name)", chore_name, str(step))
        response = self._rest.GET(url, **kwargs)
        return ChoreTask.from_dict(response.json())

    def _delete_task(self, chore_name: str, step: int, **kwargs) -> Response:
        """ Delete task from chore
        :param chore_name: name of the chore
        :param step: integer
        :return: response
        """
        url = format_url("/api/v1/Chores('{}')/Tasks({})", chore_name, str(step))
        response = self._rest.DELETE(url, **kwargs)
        return response

    def _add_task(self, chore_name: str, chore_task: ChoreTask, **kwargs) -> Response:
        """ Create Chore task on TM1 Server
        :param chore_name: name of Chore to update
        :param chore_task: instance of TM1py.ChoreTask
        :return: response
        """
        chore = self.get(chore_name, **kwargs)
        if chore.active:
            self.deactivate(chore_name, **kwargs)
        try:
            url = format_url("/api/v1/Chores('{}')/Tasks", chore_name)
            response = self._rest.POST(url, chore_task.body, **kwargs)
        except Exception as e:
            raise e
        finally:
            if chore.active:
                self.activate(chore_name, **kwargs)
        return response

    def _update_task(self, chore_name: str, chore_task: ChoreTask, **kwargs):
        """ update a chore task
        :param chore_name: name of the Chore
        :param chore_task: instance TM1py.ChoreTask
        :return: response
        """
        url = format_url("/api/v1/Chores('{}')/Tasks({})", chore_name, str(chore_task.step))
        return self._rest.PATCH(url, chore_task.body, **kwargs)

    @staticmethod
    def zfill_two(number: int) -> str:
        """ Pad an int with zeros on the left two create two digit string

        :param number:
        :return:
        """
        return str(number).zfill(2)



# -*- coding: utf-8 -*-
import json
from typing import List

from TM1py.Objects.Git import Git
from TM1py.Objects.GitCommit import GitCommit
from TM1py.Objects.GitPlan import GitPushPlan, GitPullPlan, GitPlan
from TM1py.Services.ObjectService import ObjectService
from TM1py.Services.RestService import RestService, Response
from TM1py.Utils.Utils import format_url
from TM1py.Objects.GitProject import TM1Project


class GitService(ObjectService):
    """ Service to interact with GIT
    """
    COMMON_PARAMETERS = {'username': 'Username', 'password': 'Password', 'message': 'Message', 'author': 'Author',
                         'email': 'Email', 'branch': 'Branch', 'new_branch': 'NewBranch', 'force': 'Force',
                         'public_key': 'PublicKey', 'private_key': 'PrivateKey', 'passphrase': 'Passphrase',
                         'config': 'Config'}

    def __init__(self, rest: RestService):
        super().__init__(rest)
        
    def tm1project_get(self) -> TM1Project:
        """_summary_
        """
        url = '/api/v1/!tm1project'
        response = self._rest.GET(url)
        if not response.content:
            return None

        return TM1Project.from_dict(response.json())
    
    def tm1project_delete(self):
        url = '/api/v1/!tm1project'
        empty_dict = {}
        body_json = json.dumps(empty_dict)
        
        response = self._rest.PUT(url, data=body_json)
        return TM1Project.from_dict(response.json())
        
    def tm1project_put(self, tm1_project: TM1Project) -> TM1Project:
        url = '/api/v1/!tm1project'
        body_json = tm1_project.body
        
        response = self._rest.PUT(url=url, data=body_json)
        return TM1Project.from_dict(response.json())

    def git_init(self, git_url: str, deployment: str, username: str = None, password: str = None,
                 public_key: str = None, private_key: str = None, passphrase: str = None, force: bool = None,
                 config: dict = None, **kwargs) -> Git:
        """ Initialize GIT service, returns Git object
        :param git_url: file or http(s) path to GIT repository
        :param deployment: name of selected deployment group
        :param username: GIT username
        :param password: GIT password
        :param public_key: SSH public key, available from PAA V2.0.9.4
        :param private_key: SSH private key, available from PAA V2.0.9.4
        :param passphrase: Passphrase for decrypting private key, if set
        :param force: reset git context on True
        :param config: Dictionary containing git configuration parameters
        """
        url = "/api/v1/GitInit"
        body = {'URL': git_url, 'Deployment': deployment}

        for key, value in locals().items():
            if value is not None and key in self.COMMON_PARAMETERS.keys():
                body[self.COMMON_PARAMETERS.get(key)] = value

        body_json = json.dumps(body)
        response = self._rest.POST(url=url, data=body_json, **kwargs)

        return Git.from_dict(response.json())

    def git_uninit(self, force: bool = False, **kwargs):
        """ Unitialize GIT service

        :param force: clean up git context when True
        """
        url = "/api/v1/GitUninit"
        body = json.dumps(force)
        return self._rest.POST(url=url, data=body, **kwargs)

    def git_status(self, username: str = None, password: str = None, public_key: str = None, private_key: str = None,
                   passphrase: str = None, **kwargs) -> Git:
        """ Get GIT status, returns Git object
        :param username: GIT username
        :param password: GIT password
        :param public_key: SSH public key, available from PAA V2.0.9.4
        :param private_key: SSH private key, available from PAA V2.0.9.4
        :param passphrase: Passphrase for decrypting private key, if set
        """
        url = "/api/v1/GitStatus"
        body = {}

        for key, value in locals().items():
            if value is not None and key in self.COMMON_PARAMETERS.keys():
                body[self.COMMON_PARAMETERS.get(key)] = value

        response = self._rest.POST(url=url, data=json.dumps(body), **kwargs)

        return Git.from_dict(response.json())

    def git_push(self, message: str, author: str, email: str, branch: str = None, new_branch: str = None,
                 force: bool = False, username: str = None, password: str = None, public_key: str = None,
                 private_key: str = None, passphrase: str = None, execute: bool = None, **kwargs) -> Response:
        """ Creates a gitpush plan, returns response
        :param message: Commit message
        :param author: Name of commit author
        :param email: Email of commit author
        :param branch: The branch which last commit will be used as parent commit for new branch.
        Must be empty if GIT repo is empty
        :param new_branch: If specified, creates a new branch and pushes the commit onto it. If not specified,
        pushes to the branch specified in "Branch"
        :param force: A flag passed in for evaluating preconditions
        :param username: GIT username
        :param password: GIT password
        :param public_key: SSH public key, available from PAA V2.0.9.4
        :param private_key: SSH private key, available from PAA V2.0.9.4
        :param passphrase: Passphrase for decrypting private key, if set
        :param execute: Executes the plan right away if True

        """
        url = "/api/v1/GitPush"
        body = {}

        for key, value in locals().items():
            if value is not None and key in self.COMMON_PARAMETERS.keys():
                body[self.COMMON_PARAMETERS.get(key)] = value

        response = self._rest.POST(url=url, data=json.dumps(body), **kwargs)

        if execute:
            plan_id = json.loads(response.content).get('ID')
            self.git_execute_plan(plan_id=plan_id)

        return response

    def git_pull(self, branch: str, force: bool = None, execute: bool = None, username: str = None,
                 password: str = None, public_key: str = None, private_key: str = None, passphrase: str = None,
                 **kwargs) -> Response:
        """ Creates a gitpull plan, returns response
        :param branch: The name of source branch
        :param force: A flag passed in for evaluating preconditions
        :param execute: Executes the plan right away if True
        :param username: GIT username
        :param password: GIT password
        :param public_key: SSH public key, available from PAA V2.0.9.4
        :param private_key: SSH private key, available from PAA V2.0.9.4
        :param passphrase: Passphrase for decrypting private key, if set
        """
        url = "/api/v1/GitPull"
        body = {}

        for key, value in locals().items():
            if value is not None and key in self.COMMON_PARAMETERS.keys():
                body[self.COMMON_PARAMETERS.get(key)] = value

        body_json = json.dumps(body)
        response = self._rest.POST(url=url, data=body_json, **kwargs)

        if execute:
            plan_id = json.loads(response.content).get('ID')
            self.git_execute_plan(plan_id=plan_id)

        return response

    def git_execute_plan(self, plan_id: str, **kwargs) -> Response:
        """ Executes a plan based on the planid
        :param plan_id: GitPlan id
        """
        url = format_url("/api/v1/GitPlans('{}')/tm1.Execute", plan_id)
        return self._rest.POST(url=url, **kwargs)

    def git_get_plans(self, **kwargs) -> List[GitPlan]:
        """ Gets a list of currently available GIT plans
        """
        url = "/api/v1/GitPlans"
        plans = []

        response = self._rest.GET(url=url, **kwargs)

        # Every individual plan is wrapped in a "value" parent, iterate through those to get the actual plans
        for plan in response.json().get('value'):
            plan_id = plan.get('ID')
            # Check if plan has an ID, sometimes there's a null in the mix that we don't want
            if plan_id is None:
                continue
            plan_branch = plan.get('Branch')
            plan_force = plan.get('Force')

            # A git plan can either be a PushPlan or a PullPlan, these have slightly different variables,
            # so we need to handle those differently
            if plan.get('@odata.type') == '#ibm.tm1.api.v1.GitPushPlan':
                plan_new_branch = plan.get('NewBranch')
                plan_source_files = plan.get('SourceFiles')

                new_commit = GitCommit(
                    commit_id=plan.get('NewCommit').get('ID'),
                    summary=plan.get('NewCommit').get('Summary'),
                    author=plan.get('NewCommit').get('Author'))

                parent_commit = GitCommit(
                    commit_id=plan.get('ParentCommit').get('ID'),
                    summary=plan.get('ParentCommit').get('Summary'),
                    author=plan.get('ParentCommit').get('Author'))

                current_plan = GitPushPlan(
                    plan_id=plan_id, branch=plan_branch, force=plan_force,
                    new_branch=plan_new_branch, new_commit=new_commit,
                    parent_commit=parent_commit, source_files=plan_source_files)

            elif plan.get('@odata.type') == '#ibm.tm1.api.v1.GitPullPlan':

                plan_commit = GitCommit(
                    commit_id=plan.get('Commit').get('ID'),
                    summary=plan.get('Commit').get('Summary'),
                    author=plan.get('Commit').get('Author'))

                plan_operations = plan.get('Operations')
                current_plan = GitPullPlan(plan_id=plan_id, branch=plan_branch, force=plan_force, commit=plan_commit,
                                           operations=plan_operations)

            else:
                raise RuntimeError(f"Invalid plan detected: {plan.get('@odata.type')}")

            plans.append(current_plan)

        return plans



# -*- coding: utf-8 -*-
import json
from typing import Union

from requests import Response

from TM1py.Objects.Application import DocumentApplication, ApplicationTypes, CubeApplication, ChoreApplication, \
    FolderApplication, LinkApplication, ProcessApplication, DimensionApplication, SubsetApplication, ViewApplication, \
    Application
from TM1py.Services import RestService
from TM1py.Services.ObjectService import ObjectService
from TM1py.Utils import format_url


class ApplicationService(ObjectService):
    """ Service to Read and Write TM1 Applications
    """

    def __init__(self, tm1_rest: RestService):
        """

        :param tm1_rest:
        """
        super().__init__(tm1_rest)
        self._rest = tm1_rest

    def get(self, path: str, application_type: Union[str, ApplicationTypes], name: str, private: bool = False,
            **kwargs) -> Application:
        """ Retrieve Planning Analytics Application

        :param path: path with forward slashes
        :param application_type: str or ApplicationType from Enum
        :param name:
        :param private:
        :return:
        """
        # raise ValueError if not a valid ApplicationType
        application_type = ApplicationTypes(application_type)

        # documents require special treatment
        if application_type == ApplicationTypes.DOCUMENT:
            return self.get_document(path=path, name=name, private=private, **kwargs)

        if not application_type == ApplicationTypes.FOLDER:
            name += application_type.suffix

        contents = 'PrivateContents' if private else 'Contents'
        mid = ""
        if path.strip() != '':
            mid = "".join([format_url("/Contents('{}')", element) for element in path.split('/')])

        base_url = format_url(
            "/api/v1/Contents('Applications')" + mid + "/" + contents + "('{application_name}')",
            application_name=name)

        if application_type == ApplicationTypes.CUBE:
            response = self._rest.GET(url=base_url + "?$expand=Cube($select=Name)", **kwargs)
            return CubeApplication(path=path, name=name, cube_name=response.json()["Cube"]["Name"])

        elif application_type == ApplicationTypes.CHORE:
            response = self._rest.GET(url=base_url + "?$expand=Chore($select=Name)", **kwargs)
            return ChoreApplication(path=path, name=name, chore_name=response.json()["Chore"]["Name"])

        elif application_type == ApplicationTypes.DIMENSION:
            response = self._rest.GET(url=base_url + "?$expand=Dimension($select=Name)", **kwargs)
            return DimensionApplication(path=path, name=name, dimension_name=response.json()["Dimension"]["Name"])

        elif application_type == ApplicationTypes.FOLDER:
            # implicit TM1pyException if application doesn't exist
            self._rest.GET(url=base_url, **kwargs)
            return FolderApplication(path=path, name=name)

        elif application_type == ApplicationTypes.LINK:
            # implicit TM1pyException if application doesn't exist
            self._rest.GET(url=base_url, **kwargs)
            response = self._rest.GET(base_url + "?$expand=*", **kwargs)
            return LinkApplication(path=path, name=name, url=response.json()["URL"])

        elif application_type == ApplicationTypes.PROCESS:
            response = self._rest.GET(url=base_url + "?$expand=Process($select=Name)", **kwargs)
            return ProcessApplication(path=path, name=name, process_name=response.json()["Process"]["Name"])

        elif application_type == ApplicationTypes.SUBSET:
            url = "".join([
                base_url,
                "?$expand=Subset($select=Name;$expand=Hierarchy($select=Name;$expand=Dimension($select=Name)))"])
            response = self._rest.GET(
                url=url,
                **kwargs)
            return SubsetApplication(
                path=path,
                name=name,
                dimension_name=response.json()["Subset"]["Hierarchy"]["Dimension"]["Name"],
                hierarchy_name=response.json()["Subset"]["Hierarchy"]["Name"],
                subset_name=response.json()["Subset"]["Name"])

        elif application_type == ApplicationTypes.VIEW:
            response = self._rest.GET(
                url=base_url + "?$expand=View($select=Name;$expand=Cube($select=Name))",
                **kwargs)
            return ViewApplication(
                path=path,
                name=name,
                cube_name=response.json()["View"]["Cube"]["Name"],
                view_name=response.json()["View"]["Name"])

    def get_document(self, path: str, name: str, private: bool = False, **kwargs) -> DocumentApplication:
        """ Get Excel Application from TM1 Server in binary format. Can be dumped to file.

        :param path: path through folder structure to application. For instance: "Finance/P&L.xlsx"
        :param name: name of the application
        :param private: boolean
        :return: Return DocumentApplication
        """
        if not name.endswith(ApplicationTypes.DOCUMENT.suffix):
            name += ApplicationTypes.DOCUMENT.suffix

        contents = 'PrivateContents' if private else 'Contents'
        mid = "".join([format_url("/Contents('{}')", element) for element in path.split('/')])
        url = format_url(
            "/api/v1/Contents('Applications')" + mid + "/" + contents + "('{name}')/Document/Content",
            name=name)

        content = self._rest.GET(url, **kwargs).content

        url = format_url(
            "/api/v1/Contents('Applications')" + mid + "/" + contents + "('{name}')/Document",
            name=name)
        document_fields = self._rest.GET(url, **kwargs).json()

        return DocumentApplication(
            path=path,
            name=name,
            content=content,
            file_id=document_fields.get("ID"),
            file_name=document_fields.get("Name"),
            last_updated=document_fields.get("LastUpdated"))

    def delete(self, path: str, application_type: Union[str, ApplicationTypes], application_name: str,
               private: bool = False, **kwargs) -> Response:
        """ Delete Planning Analytics application reference

        :param path: path through folder structure to delete the applications entry. For instance: "Finance/Reports"
        :param application_type: type of the to be deleted application entry
        :param application_name: name of the to be deleted application entry
        :param private: Access level of the to be deleted object
        :return:
        """

        # raise ValueError if not a valid ApplicationType
        application_type = ApplicationTypes(application_type)

        if not application_type == ApplicationTypes.FOLDER:
            application_name += application_type.suffix

        contents = 'PrivateContents' if private else 'Contents'
        mid = ""
        if path.strip() != '':
            mid = "".join([format_url("/Contents('{}')", element) for element in path.split('/')])

        url = format_url(
            "/api/v1/Contents('Applications')" + mid + "/" + contents + "('{application_name}')",
            application_name=application_name)
        return self._rest.DELETE(url, **kwargs)

    def rename(self, path: str, application_type: Union[str, ApplicationTypes], application_name: str,
               new_application_name: str, private: bool = False, **kwargs):
        # raise ValueError if not a valid ApplicationType
        application_type = ApplicationTypes(application_type)

        if not application_type == ApplicationTypes.FOLDER:
            application_name += application_type.suffix

        contents = 'PrivateContents' if private else 'Contents'
        mid = ""
        if path.strip() != '':
            mid = "".join([format_url("/Contents('{}')", element) for element in path.split('/')])

        url = format_url(
            "/api/v1/Contents('Applications')" + mid + "/" + contents + "('{application_name}')/tm1.Move",
            application_name=application_name)
        data = {"Name": new_application_name}

        return self._rest.POST(url, data=json.dumps(data), **kwargs)

    def create(self, application: Union[Application, DocumentApplication], private: bool = False, **kwargs) -> Response:
        """ Create Planning Analytics application

        :param application: instance of Application
        :param private: boolean
        :return:
        """

        contents = 'PrivateContents' if private else 'Contents'

        mid = ""
        if application.path.strip() != '':
            mid = "".join([format_url("/Contents('{}')", element) for element in application.path.split('/')])
        url = "/api/v1/Contents('Applications')" + mid + "/" + contents
        response = self._rest.POST(url, application.body, **kwargs)

        if application.application_type == ApplicationTypes.DOCUMENT:
            url = format_url(
                "/api/v1/Contents('Applications')" + mid + "/" + contents + "('{name}.blob')/Document/Content",
                name=application.name)
            response = self._rest.PUT(url, application.content, headers=self.BINARY_HTTP_HEADER, **kwargs)

        return response

    def update(self, application: Union[Application, DocumentApplication], private: bool = False, **kwargs) -> Response:
        """ Update Planning Analytics application

        :param application: instance of Application
        :param private: boolean
        :return:
        """

        contents = 'PrivateContents' if private else 'Contents'

        mid = ""
        if application.path.strip() != '':
            mid = "".join([format_url("/Contents('{}')", element) for element in application.path.split('/')])

        if application.application_type == ApplicationTypes.DOCUMENT:
            url = format_url(
                "/api/v1/Contents('Applications')" + mid + "/" + contents + "('{name}.blob')/Document/Content",
                name=application.name)
            response = self._rest.PATCH(url, application.content, headers=self.BINARY_HTTP_HEADER, **kwargs)
        else:
            url = "/api/v1/Contents('Applications')" + mid + "/" + contents
            response = self._rest.POST(url, application.body, **kwargs)
            
        return response

    def update_or_create_document_from_file(self, path: str, name: str,
                                            path_to_file: str, private: bool = False, **kwargs) -> Response:
        """Update or create application from file

        :param path: application path on server, i.e. 'Finance/Reports'
        :param name: name of the application on server, i.e. 'Flash.xlsx'
        :param path_to_file: full local file path of file, i.e. 'C:\\Users\\User\\Flash.xslx'
        :param private: access level of the object
        :return: Response
        """

        if self.exists(path=path, application_type=ApplicationTypes.DOCUMENT, name=name, private=private):
            response = self.update_document_from_file(path_to_file=path_to_file, application_path=path,
                                                    application_name=name, private=private)
        else:
            response = self.create_document_from_file(path_to_file=path_to_file, application_path=path,
                                                    application_name=name, private=private)

        return response

    def exists(self, path: str, application_type: Union[str, ApplicationTypes], name: str,
               private: bool = False, **kwargs) -> bool:
        """ Check if application exists

        :param path:
        :param application_type:
        :param name:
        :param private:
        :return:
        """
        # raise ValueError if not a valid ApplicationType
        application_type = ApplicationTypes(application_type)

        if not application_type == ApplicationTypes.FOLDER:
            name += application_type.suffix

        contents = 'PrivateContents' if private else 'Contents'
        mid = ""
        if path.strip() != '':
            mid = "".join(["/Contents('{}')".format(element) for element in path.split('/')])

        url = format_url(
            "/api/v1/Contents('Applications')" + mid + "/" + contents + "('{application_name}')",
            application_name=name)
        return self._exists(url, **kwargs)

    def create_document_from_file(self, path_to_file: str, application_path: str, application_name: str,
                                  private: bool = False, **kwargs) -> Response:
        """ Create DocumentApplication in TM1 from local file

        :param path_to_file:
        :param application_path:
        :param application_name:
        :param private:
        :return:
        """
        with open(path_to_file, 'rb') as file:
            application = DocumentApplication(path=application_path, name=application_name, content=file.read())
            return self.create(application=application, private=private, **kwargs)            

    def update_document_from_file(self, path_to_file: str, application_path: str, application_name: str,
                                  private: bool = False, **kwargs) -> Response:
        """ Update DocumentApplication in TM1 from local file

        :param path_to_file:
        :param application_path:
        :param application_name:
        :param private:
        :return:
        """
        with open(path_to_file, 'rb') as file:
            application = DocumentApplication(path=application_path, name=application_name, content=file.read())
            return self.update(application=application, private=private, **kwargs)



# -*- coding: utf-8 -*-
from typing import List

from requests import Response

from TM1py.Objects.User import User
from TM1py.Services.ObjectService import ObjectService
from TM1py.Services.RestService import RestService
from TM1py.Utils import format_url, case_and_space_insensitive_equals, require_admin


class MonitoringService(ObjectService):
    """ Service to Query and Cancel Threads in TM1
    
    """

    def __init__(self, rest: RestService):
        super().__init__(rest)

    def get_threads(self, **kwargs) -> List:
        """ Return a dict of the currently running threads from the TM1 Server

            :return:
                dict: the response
        """
        url = '/api/v1/Threads'
        response = self._rest.GET(url, **kwargs)
        return response.json()['value']

    def get_active_threads(self, **kwargs):
        """Return a list of non-idle threads from the TM1 Server

            :return:
                list: TM1 threads as dict
        """
        url = "/api/v1/Threads?$filter=Function ne 'GET /api/v1/Threads' and State ne 'Idle'"
        response = self._rest.GET(url, **kwargs)
        return response.json()['value']

    def cancel_thread(self, thread_id: int, **kwargs) -> Response:
        """ Kill a running thread
        
        :param thread_id: 
        :return: 
        """
        url = format_url("/api/v1/Threads('{}')/tm1.CancelOperation", str(thread_id))
        response = self._rest.POST(url, **kwargs)
        return response

    def cancel_all_running_threads(self, **kwargs) -> list:
        running_threads = self.get_threads(**kwargs)
        canceled_threads = list()
        for thread in running_threads:
            if thread["State"] == "Idle":
                continue
            if thread["Type"] == "System":
                continue
            if thread["Name"] == "Pseudo":
                continue
            if thread["Function"] == "GET /api/v1/Threads":
                continue
            self.cancel_thread(thread["ID"], **kwargs)
            canceled_threads.append(thread)
        return canceled_threads

    def get_active_users(self, **kwargs) -> List[User]:
        """ Get the activate users in TM1

        :return: List of TM1py.User instances
        """
        url = '/api/v1/Users?$filter=IsActive eq true&$expand=Groups'
        response = self._rest.GET(url, **kwargs)
        users = [User.from_dict(user) for user in response.json()['value']]
        return users

    def user_is_active(self, user_name: str, **kwargs) -> bool:
        """ Check if user is currently active in TM1

        :param user_name:
        :return: Boolean
        """
        url = format_url("/api/v1/Users('{}')/IsActive", user_name)
        response = self._rest.GET(url, **kwargs)
        return bool(response.json()['value'])

    def disconnect_user(self, user_name: str, **kwargs) -> Response:
        """ Disconnect User
        
        :param user_name: 
        :return: 
        """
        url = format_url("/api/v1/Users('{}')/tm1.Disconnect", user_name)
        response = self._rest.POST(url, **kwargs)
        return response

    def get_active_session_threads(self, exclude_idle: bool = True, **kwargs):
        url = "/api/v1/ActiveSession/Threads?$filter=Function ne 'GET /api/v1/ActiveSession/Threads'"
        if exclude_idle:
            url += " and State ne 'Idle'"

        response = self._rest.GET(url, **kwargs)
        return response.json()['value']

    def get_sessions(self, include_user: bool = True, include_threads: bool = True, **kwargs) -> List:
        url = "/api/v1/Sessions"
        if include_user or include_threads:
            expands = list()
            if include_user:
                expands.append("User")
            if include_threads:
                expands.append("Threads")
            url += "?$expand=" + ",".join(expands)

        response = self._rest.GET(url, **kwargs)
        return response.json()["value"]

    @require_admin
    def disconnect_all_users(self, **kwargs) -> list:
        current_user = self.get_current_user(**kwargs)
        active_users = self.get_active_users(**kwargs)
        disconnected_users = list()
        for active_user in active_users:
            if not case_and_space_insensitive_equals(current_user.name, active_user.name):
                self.disconnect_user(active_user.name, **kwargs)
                disconnected_users += [active_user.name]
        return disconnected_users

    def close_session(self, session_id, **kwargs) -> Response:
        url = format_url(f"/api/v1/Sessions('{session_id}')/tm1.Close")
        return self._rest.POST(url, **kwargs)

    @require_admin
    def close_all_sessions(self, **kwargs) -> list:
        current_user = self.get_current_user(**kwargs)
        sessions = self.get_sessions(**kwargs)
        closed_sessions = list()
        for session in sessions:
            if "User" not in session:
                continue
            if session["User"] is None:
                continue
            if "Name" not in session["User"]:
                continue
            if case_and_space_insensitive_equals(current_user.name, session["User"]["Name"]):
                continue
            self.close_session(session['ID'], **kwargs)
            closed_sessions.append(session)
        return closed_sessions

    def get_current_user(self, **kwargs):
        from TM1py import SecurityService
        security_service = SecurityService(self._rest)
        return security_service.get_current_user(**kwargs)
